from typing import Union, List, Any, Optional, TypeVar, Type
from decimal import Decimal
from otree.currency import RealWorldCurrency, Currency
from otree.database import AnyModel
from typing import Callable

class Currency(Currency):
    """
    PyCharm autocomplete seems to require that I explicitly define the class in this file
    (if I import, it says the reference to Currency is not found)
    """

cu = Currency

class DecimalUnit(Decimal):
    storage_places: int
    input_places: int
    input_unit_label: str
    output_min_places: int
    output_max_places: int

    @staticmethod
    def output(formatted, raw):
        pass

def currency_range(first, last, increment) -> List[Currency]:
    pass

from __future__ import annotations

from typing import Any, Generic, Iterable, Type, TypeVar, overload
from decimal import Decimal

# Your own types (forward-declared for the stub)
class Currency: ...
class DecimalUnit: ...

_T = TypeVar("_T")

# 2025-11-16: vscode infers instance attributes like player.x as lists apparently...
# it gives autocompletions like .append(), .pop()
# and the tooltip calls it a 'function'
# don't know why... but it's not really causing problems.

class Field(Generic[_T]):
    @overload
    def __get__(self, instance: None, owner: type[Any]) -> Field[_T]: ...
    @overload
    def __get__(self, instance: Any, owner: type[Any]) -> _T: ...
    def __set__(self, instance: Any, value: _T) -> None: ...

AnyNumber = int | float | Decimal | Currency

class models:
    def __getattr__(self, item: str) -> Any: ...

    class BooleanField(Field[bool]):
        def __init__(
            self,
            *,
            choices: Iterable | None = None,
            widget: Any | None = None,
            initial: bool | None = None,
            label: str | None = None,
            doc: str = "",
            blank: bool = False,
        ) -> None: ...

    class StringField(Field[str]):
        def __init__(
            self,
            *,
            choices: Iterable | None = None,
            widget: Any | None = None,
            initial: str | None = None,
            label: str | None = None,
            doc: str = "",
            max_length: int = 10000,
            blank: bool = False,
        ) -> None: ...

    class LongStringField(Field[str]):
        def __init__(
            self,
            *,
            initial: str | None = None,
            label: str | None = None,
            doc: str = "",
            max_length: int | None = None,
            blank: bool = False,
        ) -> None: ...

    class IntegerField(Field[int]):
        def __init__(
            self,
            *,
            choices: Iterable | None = None,
            widget: Any | None = None,
            initial: int | None = None,
            label: str | None = None,
            doc: str = "",
            min: int | None = None,
            max: int | None = None,
            blank: bool = False,
        ) -> None: ...

    class FloatField(Field[float]):
        def __init__(
            self,
            *,
            choices: Iterable | None = None,
            widget: Any | None = None,
            initial: float | None = None,
            label: str | None = None,
            doc: str = "",
            min: float | None = None,
            max: float | None = None,
            blank: bool = False,
        ) -> None: ...

    class CurrencyField(Field[Currency]):
        def __init__(
            self,
            *,
            choices: Iterable | None = None,
            widget: Any | None = None,
            initial: AnyNumber | None = None,
            label: str | None = None,
            doc: str = "",
            min: AnyNumber | None = None,
            max: AnyNumber | None = None,
            blank: bool = False,
        ) -> None: ...

    class DecimalField(Field[Decimal]):
        def __init__(
            self,
            *,
            unit: Type[DecimalUnit],
            choices: Iterable | None = None,
            widget: Any | None = None,
            initial: AnyNumber | None = None,
            label: str | None = None,
            doc: str = "",
            min: AnyNumber | None = None,
            max: AnyNumber | None = None,
            blank: bool = False,
        ) -> None: ...

    @staticmethod
    def Link(to) -> Any:
        pass

class widgets:
    def __getattr__(self, item):
        pass

    # don't need HiddenInput because you can just write <input type="hidden" ...>
    # and then you know the element's selector
    class Checkbox:
        pass

    class RadioSelect:
        pass

    class RadioSelectHorizontal:
        pass

class BaseConstants:
    pass

class BaseSubsession:
    # mark it as Any so that PyCharm can infer custom fields based on usage instead
    session: Any
    round_number: int
    def get_groups(self) -> List[GroupTV]:
        pass

    def get_group_matrix(self) -> List[List[int]]:
        pass

    def set_group_matrix(self, group_matrix: List[List[int]]):
        pass

    def get_players(self) -> List[PlayerTV]:
        pass

    def in_previous_rounds(self: SubsessionTV) -> List[SubsessionTV]:
        pass

    def in_all_rounds(self: SubsessionTV) -> List[SubsessionTV]:
        pass

    def in_round(self: SubsessionTV, round_number) -> SubsessionTV:
        pass

    def in_rounds(self: SubsessionTV, first, last) -> List[SubsessionTV]:
        pass

    def group_like_round(self, round_number: int):
        pass

    def group_randomly(self, fixed_id_in_group: bool = False):
        pass

    def field_maybe_none(self, field_name: str):
        pass

# Using TypeVar instead of the BaseSubsession seems to make PyCharm
# allow BaseSubsession be passed to a function marked as taking a Subsession arg
SubsessionTV = TypeVar("SubsessionTV", bound=BaseSubsession)

class BaseGroup:
    session: Any
    subsession: BaseSubsession
    round_number: int
    id_in_subsession: int
    def get_players(self) -> List[PlayerTV]:
        pass

    def get_player_by_role(self, role) -> PlayerTV:
        pass

    def get_player_by_id(self, id_in_group) -> PlayerTV:
        pass

    def in_previous_rounds(self: GroupTV) -> List[GroupTV]:
        pass

    def in_all_rounds(self: GroupTV) -> List[GroupTV]:
        pass

    def in_round(self: GroupTV, round_number) -> GroupTV:
        pass

    def in_rounds(self: GroupTV, first: int, last: int) -> List[GroupTV]:
        pass

    def field_maybe_none(self, field_name: str):
        pass

    def field_display(self, field_name: str):
        pass

GroupTV = TypeVar("GroupTV", bound=BaseGroup)

class BasePlayer:
    id_in_group: int
    # remove this from autocomplete because it crowds out id_in_group which is much more relevant
    # id_in_subsession: int
    payoff: Currency
    participant: Any
    session: Any
    group: GroupTV
    subsession: BaseSubsession
    round_number: int
    role: str
    def in_previous_rounds(self: PlayerTV) -> List[PlayerTV]:
        pass

    def in_all_rounds(self: PlayerTV) -> List[PlayerTV]:
        pass

    def get_others_in_group(self: PlayerTV) -> List[PlayerTV]:
        pass

    def get_others_in_subsession(self: PlayerTV) -> List[PlayerTV]:
        pass

    def in_round(self: PlayerTV, round_number) -> PlayerTV:
        pass

    def in_rounds(self: PlayerTV, first, last) -> List[PlayerTV]:
        pass

    def field_maybe_none(self, field_name: str):
        pass

    def field_display(self, field_name: str):
        pass

PlayerTV = TypeVar("PlayerTV", bound=BasePlayer)

T_extramodel = TypeVar('T_extramodel')

class ExtraModel:
    @classmethod
    def filter(cls: Type[T_extramodel], **kwargs) -> List[T_extramodel]:
        pass

    @classmethod
    def create(cls: Type[T_extramodel], **kwargs) -> T_extramodel:
        pass
    id: int

class WaitPage:
    wait_for_all_groups = False
    group_by_arrival_time = False
    title_text: str
    body_text: str
    template_name: str
    preserve_unsubmitted_inputs: bool

    round_number: int
    @staticmethod
    def is_displayed(player: Player):
        pass

    @staticmethod
    def js_vars(player: Player):
        pass

    @staticmethod
    def vars_for_template(player: Player):
        pass

    @staticmethod
    def app_after_this_page(player: Player, upcoming_apps):
        pass

    @staticmethod
    def after_all_players_arrive(group: Group):
        pass

    @staticmethod
    def live_method(player: Player, data):
        pass

class Page:
    round_number: int
    template_name: str
    timeout_seconds: int
    timer_text: str
    form_model: str
    form_fields: List[str]
    preserve_unsubmitted_inputs: bool
    allow_back_button: bool

    @staticmethod
    def live_method(player: Player, data):
        pass

    @staticmethod
    def get_form_fields(player: Player):
        pass

    @staticmethod
    def vars_for_template(player: Player):
        pass

    @staticmethod
    def js_vars(player: Player):
        pass

    @staticmethod
    def before_next_page(player: Player, timeout_happened):
        pass

    @staticmethod
    def is_displayed(player: Player):
        pass

    @staticmethod
    def error_message(player: Player, values):
        pass

    @staticmethod
    def get_timeout_seconds(player: Player):
        pass

    @staticmethod
    def app_after_this_page(player: Player, upcoming_apps):
        pass

class Bot:
    html: str
    case: Any
    cases: List
    participant: Any
    session: Any
    round_number: int
    player: PlayerTV
    group: GroupTV
    subsession: SubsessionTV

def Submission(
    PageClass, post_data: dict = {}, *, check_html=True, timeout_happened=False
):
    pass

def SubmissionMustFail(
    PageClass, post_data: dict = {}, *, check_html=True, error_fields=[]
):
    pass

def expect(*args):
    pass

# if i use Type[AnyModel] it doesn't get recognized as a superclass
# app_models = Union[Type[ExtraModel], Type[BasePlayer], Type[BaseGroup], Type[BaseSubsession]]
app_models = Union[
    Type[ExtraModel], Type[BasePlayer], Type[BaseGroup], Type[BaseSubsession]
]

def read_csv(path: str, type_model: app_models) -> List[dict]:
    pass

def render_template(path: str, *, player, C, vars: dict):
    pass

__all__ = [
    "Currency",
    "cu",
    "currency_range",
    "models",
    "widgets",
    "BaseConstants",
    "BaseSubsession",
    "BaseGroup",
    "BasePlayer",
    "ExtraModel",
    "WaitPage",
    "Page",
    "Bot",
    "Submission",
    "SubmissionMustFail",
    "expect",
    "read_csv",
    "render_template",
]



from starlette.middleware.base import BaseHTTPMiddleware
from starlette.middleware import Middleware
from starlette.middleware.sessions import SessionMiddleware
import time
from starlette.requests import Request
import logging
from otree.database import db, NEW_IDMAP_EACH_REQUEST
from otree.common import _SECRET, lock
import asyncio
import threading
from contextlib import nullcontext

logger = logging.getLogger('otree.perf')


lock2 = asyncio.Lock()

# URL paths that should NOT use the global lock
# Populated from routes after they're built to avoid duplicating URL patterns
PATHS_WITHOUT_LOCK = None


def _get_paths_without_lock():
    """Lazy load paths from routes to avoid circular import."""
    global PATHS_WITHOUT_LOCK
    if PATHS_WITHOUT_LOCK is None:
        from otree.urls import get_paths_without_lock
        PATHS_WITHOUT_LOCK = get_paths_without_lock()
    return PATHS_WITHOUT_LOCK


class CommitTransactionMiddleware(BaseHTTPMiddleware):
    async def dispatch(self, request, call_next):
        path = request.url.path

        # Check if this path should opt out of the global lock
        use_lock = not any(path.startswith(prefix) for prefix in _get_paths_without_lock())

        lock_context = lock2 if use_lock else nullcontext()

        async with lock_context:
            if NEW_IDMAP_EACH_REQUEST:
                db.new_session()
            response = await call_next(request)
            if response.status_code < 500:
                db.commit()
            else:
                db.rollback()
            return response


class PerfMiddleware(BaseHTTPMiddleware):
    async def dispatch(self, request, call_next):
        start = time.time()

        response = await call_next(request)

        # heroku has 'X-Request-ID'
        request_id = request.headers.get('X-Request-ID')
        if request_id:
            # only log this info on Heroku
            elapsed = time.time() - start
            msec = int(elapsed * 1000)
            msg = f'own_time={msec}ms request_id={request_id}'
            logger.info(msg)

        return response



import otree.views.cbv

from otree.channels import utils as channel_utils
from otree.room import ROOM_DICT, BaseRoom
from otree.session import SESSION_CONFIGS_DICT
from otree.views.admin import CreateSessionForm
from .cbv import AdminView
from .admin import mk_participant_code_urls


def get_room_shared_context(room, request):
    """Get shared context data for room views."""

    return {
        'participant_urls': room.get_participant_urls(request),
        'room_wide_url': room.get_room_wide_url(request),
        'room': room,
        'collapse_links': True,
    }


class Rooms(AdminView):
    url_pattern = '/rooms'

    def vars_for_template(self):
        from threading import get_ident

        return {'rooms': ROOM_DICT.values()}


class RoomWithoutSession(AdminView):
    room: BaseRoom
    form_class = CreateSessionForm

    url_pattern = '/room_without_session/{room_name}'

    def intercept_dispatch(self, room_name):
        self.room_name = room_name
        self.room = ROOM_DICT[room_name]
        if self.room.has_session():
            return self.redirect('RoomWithSession', room_name=room_name)

    def get_form(self):
        return CreateSessionForm(data=dict(room_name=self.room_name))

    def get_context_data(self, **kwargs):
        context = get_room_shared_context(self.room, self.request)
        context.update(
            {
                'configs': SESSION_CONFIGS_DICT.values(),
            }
        )
        ctx = super().get_context_data(**context, **kwargs)
        return ctx

    def socket_url(self):
        return channel_utils.room_admin_path(self.room.name)


class RoomWithSession(AdminView):
    template_name = 'otree/RoomWithSession.html'
    room = None

    url_pattern = '/room_with_session/{room_name}'

    def intercept_dispatch(self, room_name):
        self.room = ROOM_DICT[room_name]
        if not self.room.has_session():
            return self.redirect('RoomWithoutSession', room_name=room_name)

    def get_context_data(self, **kwargs):
        from otree.asgi import reverse

        participant_code_urls = mk_participant_code_urls(
            self.room.get_session(), self.request
        )

        session_code = self.room.get_session().code
        context = get_room_shared_context(self.room, self.request)
        context.update(
            {
                'session_url': reverse('SessionMonitor', code=session_code),
                'session_code': session_code,
                'participant_code_urls': participant_code_urls,
            }
        )
        return super().get_context_data(**context, **kwargs)


class CloseRoom(AdminView):
    url_pattern = '/CloseRoom/{room_name}'

    def post(self, request, room_name):
        self.room = ROOM_DICT[room_name]
        self.room.set_session(None)
        return self.redirect('RoomWithoutSession', room_name=room_name)



import collections
import csv
import logging
import numbers
from collections import OrderedDict
from collections import defaultdict
from html import escape
from typing import List

from sqlalchemy.sql.functions import func

import otree
from otree.common import get_models_module, inspect_field_names, get_constants
from otree.common2 import TimeSpentRow, write_page_completion_buffer
from otree.currency import Currency, RealWorldCurrency
from otree.decimal import DecimalUnit
from otree.database import dbq, values_flat
from otree.models.group import BaseGroup
from otree.models.participant import Participant
from otree.models.player import BasePlayer
from otree.models.session import Session
from otree.models.subsession import BaseSubsession
from otree.models_concrete import PageTimeBatch
from otree.session import SessionConfig
from otree import settings

logger = logging.getLogger(__name__)


# Single source of truth for fixed columns in SessionData view
# These columns always appear first, in this order
DATA_TAB_FIXED_COLUMNS = [
    {
        'field': 'participant.label',
        'display': 'label',
        'prefix': 'pp.',
        'getter': lambda p, g, s, participant_label: participant_label,
    },
    {
        'field': 'group.id_in_subsession',
        'display': 'group',
        'prefix': '',
        'getter': lambda p, g, s, participant_label: g['id_in_subsession'],
    },
]

# Automatically derive the group column index
GROUP_COL_INDEX = next(
    i
    for i, col in enumerate(DATA_TAB_FIXED_COLUMNS)
    if col['field'] == 'group.id_in_subsession'
)


def get_fields_for_monitor():
    return _get_table_fields(Participant, for_export=False)


def get_fields_for_csv(Model):
    return _get_table_fields(Model, for_export=True)


def _get_table_fields(Model, custom_fields=None, for_export=False):

    if Model is Session:
        # only data export
        return [
            'code',
            'label',
            'mturk_HITId',
            'mturk_HITGroupId',
            'comment',
            'is_demo',
        ]

    if Model is Participant:
        if for_export:
            return [
                'id_in_session',
                'code',
                'label',
                '_is_bot',
                '_index_in_pages',
                '_max_page_index',
                '_current_app_name',
                # this could be confusing because it will be in every row,
                # even rows for different rounds.
                #'_round_number',
                '_current_page_name',
                'time_started_utc',
                'visited',
                'mturk_worker_id',
                'mturk_assignment_id',
                # last so that it will be next to payoff_plus_participation_fee
                'payoff',
            ]
        else:
            return [
                'id_in_session',
                'code',
                'label',
                '_current_page_of_total',
                '_current_app_name',
                '_round_number',
                '_current_page_name',
                '_presence',
                '_monitor_note_json',
                '_last_page_timestamp',
            ]

    specs = [
        ('player', BasePlayer, ['id_in_group', 'role', 'payoff']),
        # group.id_in_subsession is in DATA_TAB_FIXED_COLUMNS, not here
        ('group', BaseGroup, []),
        # don't need round_number because it's already in the SessionData navigation toolbar
        ('subsession', BaseSubsession, []),
    ]

    for model_name_lower, BaseModel, built_in_fields in specs:
        if issubclass(Model, BaseModel):
            if custom_fields:
                fields = custom_fields.get(model_name_lower, [])
            else:
                subclass_fields = [
                    f
                    for f in inspect_field_names(Model)
                    if f not in built_in_fields
                    and f not in dir(BaseModel)
                    and not f.startswith('_')
                ]
                fields = list(dict.fromkeys(built_in_fields + subclass_fields))
            if for_export:
                return fields
            return fields


def sanitize_for_csv(value):
    if value is None:
        return ''
    if value is True:
        return 1
    if value is False:
        return 0
    if isinstance(value, (Currency, RealWorldCurrency)):
        # not decimal since that can't be json serialized
        return float(value)
    if isinstance(value, DecimalUnit):
        return float(value)
    if isinstance(value, numbers.Number):
        return value
    # Only convert to string and replace if needed
    if isinstance(value, str):
        if '\n' in value or '\r' in value:
            return value.replace('\n', ' ').replace('\r', ' ')
        return value
    return value


def tweak_player_values_dict(player: dict, group_id_in_subsession=None):
    '''because these are actually properties, the DB field starts with _.'''
    player['payoff'] = player['_payoff']
    player['role'] = player['_role']
    if group_id_in_subsession:
        player['group'] = group_id_in_subsession


_MAX_LENGTH_FOR_LIVE_UPDATE = 30


def sanitize_for_live_update(value):
    value = sanitize_for_csv(value)
    if isinstance(value, str):
        value = escape(value)
        if len(value) > _MAX_LENGTH_FOR_LIVE_UPDATE:
            value = value[:_MAX_LENGTH_FOR_LIVE_UPDATE] + '…'
    return value


def get_installed_apps_with_data() -> list:
    """
    this is just important for devserver.
    on prodserver there should never be an inconsistency between
    currently installed apps and apps with data, because you resetdb each time
    """
    app_names_with_data = set()
    for session in dbq(Session):
        for app_name in session.config['app_sequence']:
            if app_name in settings.OTREE_APPS:
                app_names_with_data.add(app_name)
    return list(sorted(app_names_with_data))


def _get_best_app_order(sessions):
    # heuristic to get the most relevant order of apps
    app_sequences = collections.Counter()
    for session in sessions:
        # we loaded the config earlier
        app_sequence = session.config['app_sequence']
        app_sequences[tuple(app_sequence)] += session.num_participants
    most_common_app_sequence = app_sequences.most_common(1)[0][0]

    app_names_with_data = get_installed_apps_with_data()

    apps_not_in_popular_sequence = [
        app for app in app_names_with_data if app not in most_common_app_sequence
    ]

    return list(most_common_app_sequence) + apps_not_in_popular_sequence


def get_rows_for_wide_csv(session_code):
    if session_code:
        sessions = [Session.objects_get(code=session_code)]
    else:
        sessions = dbq(Session).order_by('id').all()
    session_fields = get_fields_for_csv(Session)
    participant_fields = get_fields_for_csv(Participant)

    session_ids = [session.id for session in sessions]
    pps = (
        Participant.objects_filter(Participant.session_id.in_(session_ids))
        .order_by(Participant.id)
        .all()
    )
    session_cache = {row.id: row for row in sessions}

    session_config_fields = ['name']
    seen = {'name'}
    for session in sessions:
        for field_name in SessionConfig(session.config).editable_fields():
            if field_name not in seen:
                session_config_fields.append(field_name)
                seen.add(field_name)

    if not pps:
        # 1 empty row
        return [[]]

    header_row = [f'participant.{fname}' for fname in participant_fields]
    header_row += [f'participant.{fname}' for fname in settings.PARTICIPANT_FIELDS]
    header_row += [f'session.{fname}' for fname in session_fields]
    header_row += [f'session.config.{fname}' for fname in session_config_fields]
    header_row += [f'session.{fname}' for fname in settings.SESSION_FIELDS]
    rows = [header_row]

    for pp in pps:
        session = session_cache[pp.session_id]
        row = [getattr(pp, fname) for fname in participant_fields]
        row += [pp.vars.get(fname, None) for fname in settings.PARTICIPANT_FIELDS]
        row += [getattr(session, fname) for fname in session_fields]
        row += [session.config.get(fname) for fname in session_config_fields]
        row += [session.vars.get(fname, None) for fname in settings.SESSION_FIELDS]
        rows.append(row)

    order_of_apps = _get_best_app_order(sessions)

    rounds_per_app = OrderedDict()
    for app_name in order_of_apps:
        try:
            models_module = get_models_module(app_name)
        except ModuleNotFoundError:
            # this should only happen with devserver because on production server,
            # you would need to resetdb after renaming an app.
            logger.warning(
                f'Cannot export data for app {app_name}, which existed when the session was run '
                f'but no longer exists.'
            )
            continue

        highest_round_number = dbq(
            func.max(models_module.Subsession.round_number)
        ).scalar()

        if highest_round_number is not None:
            rounds_per_app[app_name] = highest_round_number

    for app_name in rounds_per_app:
        app_rows = get_rows_for_wide_csv_app(
            app_name, rounds_per_app[app_name], sessions
        )
        for i in range(len(rows)):
            rows[i].extend(app_rows[i])

    return [[sanitize_for_csv(v) for v in row] for row in rows]


def get_rows_for_wide_csv_app(app_name, max_round_number, sessions: List[Session]):
    """Optimized version that loads all rounds for an app at once"""
    models_module = otree.common.get_models_module(app_name)
    Player: BasePlayer = models_module.Player
    Group: BaseGroup = models_module.Group
    Subsession: BaseSubsession = models_module.Subsession
    pfields = get_fields_for_csv(Player)
    gfields = get_fields_for_csv(Group)
    sfields = get_fields_for_csv(Subsession)

    # Load ALL data for this app at once
    all_groups = Group.values_dicts()
    groups_by_round = defaultdict(dict)
    for g in all_groups:
        groups_by_round[g['round_number']][g['id']] = g

    session_ids = [s.id for s in sessions]
    all_subsessions = Subsession.values_dicts()
    subsessions_by_session_round = {}
    for s in all_subsessions:
        subsessions_by_session_round[(s['session_id'], s['round_number'])] = s

    all_players = Player.values_dicts(order_by='id')
    players_by_subsession = defaultdict(list)
    for p in all_players:
        players_by_subsession[p['subsession_id']].append(p)

    # Now build rows for each round
    all_app_rows = []
    for round_number in range(1, max_round_number + 1):
        rows = []
        group_cache = groups_by_round.get(round_number, {})

        # Build header
        header_row = []
        for model_name, fields in [
            ('player', pfields),
            ('group', gfields),
            ('subsession', sfields),
        ]:
            for fname in fields:
                header_row.append(f'{app_name}.{round_number}.{model_name}.{fname}')
        rows.append(header_row)
        empty_row = ['' for _ in range(len(header_row))]

        # Process each session
        for session in sessions:
            subsession = subsessions_by_session_round.get((session.id, round_number))
            if not subsession:
                subsession_rows = [empty_row for _ in range(session.num_participants)]
            else:
                players = players_by_subsession.get(subsession['id'], [])

                if len(players) != session.num_participants:
                    msg = (
                        f"Session {session.code} has {session.num_participants} participants, "
                        f"but round {round_number} of app '{app_name}' "
                        f"has {len(players)} players. The number of players in the subsession "
                        "should always match the number of players in the session. "
                        "Please report this issue and then reset the database."
                    )
                    raise AssertionError(msg)

                subsession_rows = []
                for player in players:
                    group = group_cache[player['group_id']]
                    tweak_player_values_dict(player)

                    row = [player[fname] for fname in pfields]
                    row += [group[fname] for fname in gfields]
                    row += [subsession[fname] for fname in sfields]

                    subsession_rows.append(row)
            rows.extend(subsession_rows)
        all_app_rows.append(rows)

    # Transpose: convert from list of (rounds × rows) to list of (rows × rounds)
    # all_app_rows structure: [[round1_header, round1_data_row1, round1_data_row2, ...], [round2_header, round2_data_row1, ...], ...]
    # We want: [all_headers_combined, all_data_row1_combined, all_data_row2_combined, ...]
    if not all_app_rows:
        return []

    num_rows = len(all_app_rows[0])
    transposed = [[] for _ in range(num_rows)]

    for round_rows in all_app_rows:
        # Each round_rows has: [header, data_row_1, data_row_2, ...]
        # We skip the header (index 0) for data rows, only concatenate headers at index 0
        for i in range(len(round_rows)):
            if i == 0:
                # This is a header row - extend with this round's header
                transposed[0].extend(round_rows[0])
            else:
                # This is a data row - extend with this round's data (skip header)
                transposed[i].extend(round_rows[i])

    return transposed


def get_rows_for_csv(app_name, session_code=None):
    # need to use app_name and not app_label because the app might have been
    # removed from SESSION_CONFIGS
    models_module = otree.common.get_models_module(app_name)
    Player = models_module.Player
    Group = models_module.Group
    Subsession = models_module.Subsession

    columns_for_models = {
        Model.__name__.lower(): get_fields_for_csv(Model)
        for Model in [Player, Group, Subsession, Participant, Session]
    }

    if session_code:
        session_ids = [Session.objects_get(code=session_code).id]
    else:
        session_ids = values_flat(dbq(Subsession), Subsession.session_id)

    # Helper to filter by session
    def values_for_session(Model):
        if session_ids:
            return Model.values_dicts(Model.session_id.in_(session_ids))
        return Model.values_dicts()

    players = list(values_for_session(Player))
    # Sort after filtering to maintain order
    players.sort(key=lambda p: p['id'])

    value_dicts = dict(
        group={row['id']: row for row in values_for_session(Group)},
        subsession={row['id']: row for row in values_for_session(Subsession)},
        participant={row['id']: row for row in values_for_session(Participant)},
        session={
            row['id']: row for row in Session.values_dicts(Session.id.in_(session_ids))
        },
    )

    model_order = ['participant', 'player', 'group', 'subsession', 'session']

    # header row
    rows = [[f'{m}.{col}' for m in model_order for col in columns_for_models[m]]]

    for player in players:
        tweak_player_values_dict(player)
        row = []

        for model_name in model_order:
            if model_name == 'player':
                obj = player
            else:
                obj = value_dicts[model_name][player[f'{model_name}_id']]
            for colname in columns_for_models[model_name]:
                row.append(sanitize_for_csv(obj[colname]))
        rows.append(row)

    return rows


def get_rows_for_monitor(participants) -> list:
    field_names = get_fields_for_monitor()
    callable_fields = {'_numeric_label', '_current_page_of_total'}
    rows = []
    for pp in participants:
        row = {}
        for field_name in field_names:
            value = getattr(pp, field_name)
            if field_name in callable_fields:
                value = value()
            row[field_name] = value
        row['id_in_session'] = pp.id_in_session
        row['status'] = pp.status
        if pp.is_on_wait_page:
            if pp._waitpage_is_connected:
                if pp._waitpage_tab_hidden:
                    icon = '🟡'
                else:
                    icon = '🟢'
            else:
                icon = '⚪'
            row['_presence'] = icon
        rows.append(row)
    return rows


def get_rows_for_data_tab(session):
    for app_name in session.config['app_sequence']:
        yield from get_rows_for_data_tab_app(session, app_name)


def get_fields_for_data_tab(app_name):
    C = get_constants(app_name)
    custom_fields = getattr(C, ADMIN_VIEW_FIELDS, {})
    parse_custom_fields(custom_fields)
    models_module = get_models_module(app_name)
    for Model in [models_module.Player, models_module.Group, models_module.Subsession]:
        yield _get_table_fields(Model, for_export=False, custom_fields=custom_fields)


def parse_custom_fields(custom_fields):
    for k in custom_fields:
        if k not in ['player', 'group', 'subsession']:
            msg = f"{ADMIN_VIEW_FIELDS} dict contains invalid key: {k}. "
            raise Exception(msg)
    return custom_fields


ADMIN_VIEW_FIELDS = 'ADMIN_VIEW_FIELDS'


def get_rows_for_data_tab_app(session, app_name):
    models_module = get_models_module(app_name)
    Player = models_module.Player
    Group = models_module.Group
    Subsession = models_module.Subsession
    C = get_constants(app_name)
    custom_fields = getattr(C, ADMIN_VIEW_FIELDS, None)
    pfields, gfields, sfields = get_fields_for_data_tab(app_name)

    players = Player.values_dicts(session=session, order_by='id')

    # Load just participant id and label
    participant_labels = {
        p_id: label
        for p_id, label in dbq(Participant)
        .filter(Participant.session_id == session.id)
        .with_entities(Participant.id, Participant.label)
        .all()
    }

    players_by_round = defaultdict(list)
    for p in players:
        players_by_round[p['round_number']].append(p)

    groups = {g['id']: g for g in Group.values_dicts(session=session)}
    subsessions = {s['id']: s for s in Subsession.values_dicts(session=session)}

    for round_number in range(1, len(subsessions) + 1):
        table = []
        for p in players_by_round[round_number]:
            participant_label = participant_labels[p['participant_id']]
            g = groups[p['group_id']]
            tweak_player_values_dict(p, g['id_in_subsession'])
            s = subsessions[p['subsession_id']]
            try:
                row = (
                    [
                        col['getter'](p, g, s, participant_label)
                        for col in DATA_TAB_FIXED_COLUMNS
                    ]
                    + [p[fname] for fname in pfields]
                    + [g[fname] for fname in gfields]
                    + [s[fname] for fname in sfields]
                )
            except KeyError as exc:
                if custom_fields:
                    msg = f"{ADMIN_VIEW_FIELDS}: field {repr(exc.args[0])}"
                    raise Exception(msg) from None
                raise exc
            table.append([sanitize_for_live_update(v) for v in row])
        yield table


def export_wide(fp, session_code=None):
    rows = get_rows_for_wide_csv(session_code=session_code)
    _export_csv(fp, rows)


def export_app(app_name, fp, session_code=None):
    rows = get_rows_for_csv(app_name, session_code=session_code)
    _export_csv(fp, rows)


from sqlalchemy.orm import joinedload


def get_custom_export_functions(app_name):
    """Get all custom export functions from an app's models module.

    Returns a dict mapping function names to functions.
    Functions that start with 'custom_export' are included.
    """
    models_module = get_models_module(app_name)
    custom_exports = {}

    for attr_name in dir(models_module):
        if attr_name.startswith('custom_export'):
            attr = getattr(models_module, attr_name)
            if callable(attr):
                custom_exports[attr_name] = attr

    return custom_exports


def custom_export_app(app_name, fp, session_code=None, function_name='custom_export'):

    models_module = get_models_module(app_name)
    Player = models_module.Player
    query = dbq(Player).order_by('id')

    if session_code:
        session = Session.objects_get(code=session_code)
        query = query.filter(Player.session_id == session.id)

    qs = list(
        query.options(
            joinedload(Player.participant, innerjoin=True),
            joinedload(Player.group, innerjoin=True),
            joinedload(Player.subsession, innerjoin=True),
            joinedload(Player.session, innerjoin=True),
        )
    )
    for player in qs:
        # need this to query null values
        player._is_frozen = False

    # Get the specific export function
    export_func = getattr(models_module, function_name, None)
    if not export_func:
        raise ValueError(
            f"Export function '{function_name}' not found in app '{app_name}'"
        )

    rows = export_func(qs)
    str_rows = []
    for row in rows:
        str_rows.append([sanitize_for_csv(ele) for ele in row])
    _export_csv(fp, str_rows)


def _export_csv(fp, rows):
    writer = csv.writer(fp)
    writer.writerows(rows)


def export_page_times(fp):
    write_page_completion_buffer()
    batches = values_flat(dbq(PageTimeBatch).order_by('id'), PageTimeBatch.text)
    fp.write(','.join(TimeSpentRow.__annotations__.keys()) + '\n')
    for batch in batches:
        fp.write(batch)


def export_chat(fp, session_code=None):
    from otree.models_concrete import ChatMessage

    column_names = [
        'session_code',
        'id_in_session',
        'participant_code',
        'channel',
        'nickname',
        'body',
        'timestamp',
    ]

    query = dbq(ChatMessage).join(Participant).order_by(ChatMessage.timestamp)

    if session_code:
        query = query.filter(Participant._session_code == session_code)

    rows = query.with_entities(
        Participant._session_code,
        Participant.id_in_session,
        Participant.code,
        ChatMessage.channel,
        ChatMessage.nickname,
        ChatMessage.body,
        ChatMessage.timestamp,
    )

    writer = csv.writer(fp)
    writer.writerows([column_names])
    writer.writerows(rows)


BOM = '\ufeff'



import os
import sys
import otree
from .base import BaseCommand
import shutil
from pathlib import Path
from tempfile import TemporaryDirectory
from io import BytesIO
import random

print_function = print


def prompt_about_sample_games():
    '''for easy patching'''
    return input("Include sample games? (y or n): ")


class Command(BaseCommand):
    help = "Creates a new oTree project."

    def add_arguments(self, parser):
        parser.add_argument('name')
        # note-to-self
        # we need a CLI arg rather than a small function we can patch,
        # because our test launches this in a subprocess.
        parser.add_argument(
            '--noinput',
            action='store_false',
            dest='interactive',
            default=True,
        )

    def handle(self, name, interactive):
        dest = Path(name)
        if Path('settings.py').exists():
            msg = (
                'You are trying to create a project but it seems you are '
                'already in a project folder (found settings.py).'
            )
            sys.exit(msg)
        if dest.exists():
            msg = (
                f'There is already a project called "{name}" '
                'in this folder. Either delete that folder first, or use a different name.'
            )
            sys.exit(msg)

        if interactive and prompt_about_sample_games().lower() == "y":
            download_from_github(dest)
        else:
            copy_project_template(dest)
        settings_path = dest.joinpath('settings.py')
        settings_path.write_text(
            settings_path.read_text().replace(
                "{{ secret_key }}", str(random.randint(10**12, 10**13))
            )
        )

        msg = (
            'Created project folder.\n'
            f'Enter "cd {name}" to move inside the project folder, '
            'then start the server with "otree devserver".'  #
        )
        print_function(msg)


def download_from_github(dest: Path):
    # expensive import
    from urllib.request import urlopen
    import zipfile

    branch_name = 'lite'
    resp = urlopen(f"https://github.com/oTree-org/oTree/archive/{branch_name}.zip")
    f = BytesIO()
    f.write(resp.read())
    f.seek(0)
    with TemporaryDirectory() as tmpdir:
        with zipfile.ZipFile(f, 'r') as zip_ref:
            # omit tests.py because it is jarring/distracting with __init__.py format.
            zip_ref.extractall(
                tmpdir,
                members=[f for f in zip_ref.namelist() if not f.endswith('tests.py')],
            )
        shutil.move(Path(tmpdir, f'oTree-{branch_name}'), dest)


def copy_project_template(dest: Path):
    src = Path(otree.__file__).parent / 'assets/project_template'
    shutil.copytree(src, dest)



import logging
import os
import sys
from logging.config import dictConfig
from pathlib import Path
from sys import argv

from otree import __version__

# adapted from uvicorn
LOGGING_CONFIG = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "default": {
            # we could use colorama to get coloring for windows. otherwise we get:
            # [32mINFO[0m:
            # "()": "uvicorn.logging.DefaultFormatter",
            "fmt": "%(levelprefix)s %(message)s",
            "use_colors": None,
        },
        "access": {
            "()": "uvicorn.logging.AccessFormatter",
            "fmt": '%(levelprefix)s %(client_addr)s - "%(request_line)s" %(status_code)s',  # noqa: E501
        },
    },
    "handlers": {
        "default": {
            "formatter": "default",
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stderr",
        },
        "access": {
            "formatter": "access",
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stdout",
        },
    },
    "loggers": {
        "uvicorn": {"handlers": ["default"], "level": "INFO"},
        "uvicorn.error": {"level": "INFO", "propagate": False},
        "uvicorn.access": {"handlers": ["access"], "level": "INFO", "propagate": False},
        # otherwise we get 'connection open', 'connection closed' on each page
        "websockets.server": {"propagate": False, "level": "WARNING"},
    },
    'root': {
        'handlers': ['default'],
        'level': logging.INFO,
    },
}

dictConfig(LOGGING_CONFIG)

logger = logging.getLogger(__name__)


print_function = print

COMMAND_ALIASES = dict(
    test='bots',
    runprodserver='prodserver',
    webandworkers='prodserver1of2',
    runprodserver1of2='prodserver1of2',
    runprodserver2of2='prodserver2of2',
    timeoutworker='prodserver2of2',
    version='--version',
    help='--help',
)


def execute_from_command_line(*args, **kwargs):

    if len(argv) == 1:
        argv.append('help')

    cmd = argv[1]
    cmd = COMMAND_ALIASES.get(cmd, cmd)

    if cmd == '--version':
        print_function(__version__)
        return
    if cmd == '--help':
        print_function(MAIN_HELP_TEXT)
        return

    # [does the below caveat still apply without django?]
    # need to set env var rather than setting otree.common.USE_TIMEOUT_WORKER because
    # that module cannot be loaded yet.
    if cmd in ['prodserver', 'prodserver1of2']:
        os.environ['USE_TIMEOUT_WORKER'] = '1'
    if 'devserver' in cmd:
        os.environ['USE_TEMPLATE_RELOADER'] = '1'

    if cmd in [
        'startproject',
        'version',
        '--version',
        'unzip',
        'zip',
        'zipserver',
        'devserver',
    ]:
        # skip full setup.
        pass
    else:
        if cmd in ['devserver_inner', 'bots']:
            os.environ['OTREE_IN_MEMORY'] = '1'
            # note-to-self
            # 2023-07-11: seems that bots should start from scratch,
            # so that when the bot runner exports CSV, it should only have bot data.
            # however, i got a sqlalchemy error that suggests I need to do something
            # with DBWrapper and new_connection. Look into it later.
            # os.environ['OTREE_EPHEMERAL'] = '1'
        setup()

    from otree.cli.base import call_command

    call_command(cmd, *argv[2:])


def setup():

    from otree import settings

    init_i18n(settings.LANGUAGE_CODE_ISO)

    # Import CURRENCY_UNIT class if defined (delayed to avoid circular imports)
    if settings.CURRENCY_UNIT_PATH:
        import importlib
        import sys
        from otree.decimal import DecimalUnit

        try:
            module_path, class_name = settings.CURRENCY_UNIT_PATH.rsplit('.', 1)
            module = importlib.import_module(module_path)
            currency_unit_class = getattr(module, class_name)

            # Validate it's a DecimalUnit subclass
            if not issubclass(currency_unit_class, DecimalUnit):
                msg = (
                    f"CURRENCY_UNIT must be a subclass of DecimalUnit, "
                    f"but {settings.CURRENCY_UNIT_PATH} is not."
                )
                sys.exit(msg)

            # Set it globally - Currency/RealWorldCurrency __new__ will use this
            settings.CURRENCY_UNIT_CLASS = currency_unit_class

        except (ValueError, ImportError, AttributeError) as e:
            msg = f"Failed to import CURRENCY_UNIT '{settings.CURRENCY_UNIT_PATH}': {e}"
            sys.exit(msg)

    from otree.database import init_orm  # noqa

    check_for_sentry()
    init_orm()

    import otree.bots.browser

    otree.bots.browser.browser_bot_worker = otree.bots.browser.BotWorker()


def init_i18n(LANGUAGE_CODE_ISO):
    os.environ['LANGUAGE'] = LANGUAGE_CODE_ISO
    import gettext

    # note-to-self
    # because the files are called django.mo
    gettext.textdomain('django')
    gettext.bindtextdomain('django', localedir=str(Path(__file__).parent / 'locale'))
    if Path('_locale').is_dir():
        gettext.bindtextdomain('messages', localedir='_locale')
        gettext.textdomain('messages')


def send_termination_notice(PORT) -> int:
    # note-to-self
    # expensive imports. do it in here so that we don't delay devserver_inner
    from urllib.request import urlopen
    import urllib.error

    try:
        # note-to-self
        # send data= so it makes a post request
        # it seems using localhost is very slow compared to 127.0.0.1
        resp = urlopen(f'http://127.0.0.1:{PORT}/SaveDB', data=b'foo')
        return int(resp.read().decode('utf-8'))
    except urllib.error.URLError as exc:
        # note-to-self
        # - URLError may happen if the server didn't even start up yet
        #  (if you stop it right away)
        pass


def check_for_sentry():
    SENTRY_DSN = os.environ.get('SENTRY_DSN')
    if SENTRY_DSN:
        try:
            import sentry_sdk
        except ModuleNotFoundError:
            sys.exit(
                'For Sentry to work, you need to add sentry_sdk to your requirements.txt.'
            )
        from sentry_sdk.utils import json_dumps
        from urllib.request import urlopen
        import urllib.error

        def before_send(event, hint):
            otree_post_url = os.getenv('OTREE_ERROR_REPORT_ENDPOINT')
            if otree_post_url:
                try:
                    urlopen(otree_post_url, data=json_dumps(event), timeout=0.5)
                except urllib.error.URLError:
                    # in case timeout is raised etc
                    pass
            return event

        sentry_sdk.init(
            dsn=SENTRY_DSN,
            # note-to-self
            # 2018-11-24: breadcrumbs were causing memory leaks when doing queries,
            # especially when creating sessions, which construct hugely verbose
            # queries with bulk_create.
            # however, i could only clearly observe the difference this line makes
            # when testing
            # on a script that bulk_created thousands of non-otree models.
            # when testing on a live server, things are more ambiguous.
            # maybe just refreshing the page several times after creating a session
            # is enough to reset memory to reasnoable levels?
            # disabling also may make things faster...
            # in anecdotal test, 40 vs 50 seconds
            max_breadcrumbs=0,
            release=__version__,
            before_send=before_send,
        )


MAIN_HELP_TEXT = '''
Available subcommands:

browser_bots
create_session
devserver
prodserver
prodserver1of2
prodserver2of2
resetdb
startapp
startapp_trials
startproject
test
unzip
zip
zipserver
'''



from decimal import Decimal, ROUND_HALF_UP
from otree.i18n import convert_decimal_separator


def _to_decimal(amount):
    if isinstance(amount, Decimal):
        return amount
    elif isinstance(amount, float):
        return Decimal.from_float(amount)
    else:
        return Decimal(amount)


class DecimalUnit(Decimal):

    # Default class attributes (should be overridden by subclasses)
    storage_places = None
    input_places = None
    input_unit_label = ''
    output_min_places = None
    output_max_places = None

    @staticmethod
    def output(formatted: str, raw: Decimal) -> str:
        return formatted

    def __new__(cls, amount):
        """Create an instance of this decimal unit"""
        if amount is None:
            raise ValueError('Cannot convert None to decimal')
        cls._validate_config()
        instance = Decimal.__new__(cls, cls._sanitize(amount))
        return instance

    @classmethod
    def _validate_config(cls):
        """Validate that required class attributes are defined"""
        if cls.storage_places is None:
            raise ValueError(
                f'{cls.__name__} is missing required attribute "storage_places"'
            )
        if cls.input_places is None:
            raise ValueError(
                f'{cls.__name__} is missing required attribute "input_places"'
            )
        if cls.output_min_places is None:
            raise ValueError(
                f'{cls.__name__} is missing required attribute "output_min_places"'
            )
        if cls.output_max_places is None:
            raise ValueError(
                f'{cls.__name__} is missing required attribute "output_max_places"'
            )

    @classmethod
    def _sanitize(cls, amount):
        if isinstance(amount, cls):
            return amount
        places = cls.storage_places
        quant = Decimal('0.1') ** places
        return _to_decimal(amount).quantize(quant, rounding=ROUND_HALF_UP)

    # Support for pickling
    def __reduce__(self):
        return (self.__class__, (Decimal.__str__(self),))

    # Immutable
    def __copy__(self):
        return self

    def __deepcopy__(self, memo):
        return self

    def __float__(self):
        return float(Decimal(self))

    def __str__(self):
        return self._format_for_display()

    def __repr__(self):
        # don't use .normalize() because:
        # (1) easier to understand the precision
        # (2) numbers line up, e.g. in data table view.
        return f'{self.__class__.__name__}({Decimal(self)})'

    def __eq__(self, other):
        if isinstance(other, DecimalUnit):
            # Only equal if same type and same value
            return type(self) == type(other) and Decimal.__eq__(self, other)
        elif isinstance(other, (int, float, Decimal)):
            return Decimal.__eq__(self, self._sanitize(other))
        else:
            return False

    __hash__ = Decimal.__hash__

    def _format_with_places(self, min_places, max_places):
        """Helper to format with specific decimal places and apply custom function"""
        # Round to max_places
        quant = Decimal('0.1') ** max_places
        rounded = Decimal(self).quantize(quant, rounding=ROUND_HALF_UP)

        # Format with Python's thousand separator (comma) and max_places decimals
        fmt = "{:,.%df}" % max_places
        result = fmt.format(rounded)

        # Strip trailing zeros down to min_places (before converting separators)
        if min_places < max_places and '.' in result:
            parts = result.split('.')
            decimal_part = parts[1].rstrip('0')
            # Ensure we keep at least min_places
            decimal_part = decimal_part.ljust(min_places, '0')
            if decimal_part:
                result = parts[0] + '.' + decimal_part
            else:
                result = parts[0]

        # Convert ',' to THOUSAND_SEPARATOR and '.' to DECIMAL_SEPARATOR
        result = convert_decimal_separator(result)

        return self.__class__.output(result, Decimal(self))

    def _prepare_operand(self, other):
        """Convert operand to Decimal for arithmetic"""
        if isinstance(other, Decimal):
            return other
        return _to_decimal(other)

    @staticmethod
    def _make_binary_operator(name):
        """Factory for creating arithmetic operator methods"""
        method = getattr(Decimal, name, None)

        def binary_function(self, other):
            other = self._prepare_operand(other)
            result = method(self, other)
            return self._make_result(result, other)

        return binary_function

    def _format_for_display(self):
        """Format using unit's display settings"""
        min_places = self.__class__.output_min_places
        max_places = self.__class__.output_max_places
        return self._format_with_places(min_places, max_places)

    def to_real_world_currency(self, session):
        """
        Compatibility method for when CurrencyField is aliased to DecimalField.
        Since there's only one currency type when CURRENCY_UNIT is defined,
        just return self (similar to Currency.to_real_world_currency when USE_POINTS=False)
        """
        return self

    def _make_result(self, result, other):
        """DRY helper for arithmetic operations"""
        if not isinstance(other, DecimalUnit):
            # Operating with scalar preserves type
            return self.__class__(result)
        if type(self) == type(other):
            # Same type preserves type
            return self.__class__(result)
        # Different types return plain Decimal
        return Decimal(result)

    # Arithmetic operations - generated by _make_binary_operator
    __add__ = _make_binary_operator('__add__')
    __radd__ = _make_binary_operator('__radd__')
    __sub__ = _make_binary_operator('__sub__')
    __rsub__ = _make_binary_operator('__rsub__')
    __mul__ = _make_binary_operator('__mul__')
    __rmul__ = _make_binary_operator('__rmul__')
    __truediv__ = _make_binary_operator('__truediv__')
    __rtruediv__ = _make_binary_operator('__rtruediv__')
    __floordiv__ = _make_binary_operator('__floordiv__')
    __rfloordiv__ = _make_binary_operator('__rfloordiv__')
    __mod__ = _make_binary_operator('__mod__')
    __rmod__ = _make_binary_operator('__rmod__')
    __pow__ = _make_binary_operator('__pow__')
    __rpow__ = _make_binary_operator('__rpow__')

    def __neg__(self):
        return self.__class__(Decimal.__neg__(self))

    def __pos__(self):
        return self.__class__(Decimal.__pos__(self))

    def __abs__(self):
        return self.__class__(Decimal.__abs__(self))



from . import nodes
from . import errors


# Token delimiters.
COMMENT_START = '{#'
COMMENT_END = '#}'
INSTRUCTION_START_OLD = '{%'
INSTRUCTION_END_OLD = '%}'
INSTRUCTION_START = '{{'
INSTRUCTION_END = '}}'


# Returns the root node of the compiled node tree.
def compile(template_string, template_id, is_page_template):
    return Parser(
        template_string, template_id, is_page_template=is_page_template
    ).parse()


# Tokens come in four different types: TEXT, PRINT, EPRINT, and INSTRUCTION.
class Token:
    def __init__(self, token_type, token_text, template_id, line_number):
        words = token_text.split()
        self.keyword = words[0] if words else ''
        self.type = token_type
        self.text = token_text
        self.template_id = template_id
        self.line_number = line_number

    def __str__(self):
        return (
            f"({self.type}, {repr(self.text)}, {self.template_id}, {self.line_number})"
        )


# The Lexer takes a template string as input and chops it into a list of Tokens.
class Lexer:
    def __init__(self, template_string, template_id):
        self.template_string = template_string
        self.template_id = template_id
        self.tokens = []
        self.index = 0
        self.line_number = 1

    def tokenize(self):
        while self.index < len(self.template_string):
            if self.match(COMMENT_START):
                self.read_comment_tag()
            elif self.match(INSTRUCTION_START) or self.match(INSTRUCTION_START_OLD):
                self.read_instruction_tag()
            else:
                self.read_text()
        return self.tokens

    def match(self, target):
        if self.template_string.startswith(target, self.index):
            return True
        return False

    def advance(self):
        if self.template_string[self.index] == '\n':
            self.line_number += 1
        self.index += 1

    def read_comment_tag(self):
        self.index += len(COMMENT_START)
        start_line_number = self.line_number
        while self.index < len(self.template_string):
            if self.match(COMMENT_END):
                self.index += len(COMMENT_END)
                return
            self.advance()
        msg = f"Unclosed comment tag"
        raise errors.TemplateLexingError(msg, self.template_id, start_line_number)

    def read_instruction_tag(self):
        self.index += len(INSTRUCTION_START)
        start_index = self.index
        start_line_number = self.line_number
        while self.index < len(self.template_string):
            if self.match(INSTRUCTION_END) or self.match(INSTRUCTION_END_OLD):
                text = self.template_string[start_index : self.index].strip()
                self.tokens.append(
                    Token("INSTRUCTION", text, self.template_id, start_line_number)
                )
                self.index += len(INSTRUCTION_END)
                return
            self.advance()
        msg = f"Unclosed instruction tag"
        raise errors.TemplateLexingError(msg, self.template_id, start_line_number)

    def read_text(self):
        start_index = self.index
        start_line_number = self.line_number
        while self.index < len(self.template_string):
            if self.match(COMMENT_START):
                break
            elif self.match(INSTRUCTION_START) or self.match(INSTRUCTION_START_OLD):
                break
            self.advance()
        text = self.template_string[start_index : self.index]
        self.tokens.append(Token("TEXT", text, self.template_id, start_line_number))


# The Parser takes a template string as input, lexes it into a token stream, then compiles the
# token stream into a tree of nodes.
class Parser:
    def __init__(self, template_string, template_id, is_page_template):
        self.template_string = template_string
        self.template_id = template_id
        self.num_nested_blocks = 0
        self.is_page_template = is_page_template

    def parse(self):
        stack = [nodes.Node()]
        expecting = []

        for token in Lexer(self.template_string, self.template_id).tokenize():
            if (
                self.is_page_template
                and self.num_nested_blocks == 0
                and token.type == 'TEXT'
                and token.text.strip()
            ):
                # technically the 'text' instruction also includes leading whitespace
                # which is really confusing because it looks like it
                # points to the previous instruction tag
                num_leading_whitespace = len(token.text) - len(token.text.lstrip())
                num_leading_newlines = token.text[:num_leading_whitespace].count('\n')
                token.line_number += num_leading_newlines
                raise errors.TemplateSyntaxError(
                    "All content must be inside a block tag", token
                )
            if token.type == "TEXT":
                stack[-1].children.append(nodes.TextNode(token))
            elif token.keyword in nodes.instruction_keywords:
                if token.keyword == 'block':
                    self.num_nested_blocks += 1
                node_class, endword = nodes.instruction_keywords[token.keyword]
                node = node_class(token)
                stack[-1].children.append(node)
                if endword:
                    stack.append(node)
                    expecting.append(endword)
            elif token.keyword in nodes.instruction_endwords:
                if len(expecting) == 0:
                    msg = f"Unexpected tag"
                    raise errors.TemplateSyntaxError(msg, token)
                elif expecting[-1] != token.keyword:
                    msg = (
                        f"Unexpected '{token.keyword}' tag. "
                        f"Was expecting the following closing tag: '{expecting[-1]}'."
                    )
                    raise errors.TemplateSyntaxError(msg, token)
                else:
                    stack[-1].exit_scope()
                    stack.pop()
                    expecting.pop()
                    if token.keyword == 'endblock':
                        self.num_nested_blocks -= 1
            elif token.keyword == '':
                msg = f"Empty instruction tag"
                raise errors.TemplateSyntaxError(msg, token)
            else:
                stack[-1].children.append(nodes.PrintNode(token))

        if expecting:
            token = stack[-1].token
            msg = (
                f"Unexpected end of template. "
                f"Was expecting a closing tag '{expecting[-1]}' to close the "
                f"'{token.keyword}' tag opened in line {token.line_number}."
            )
            raise errors.TemplateSyntaxError(msg, token)

        return stack.pop()



import math
import json
from starlette.background import BackgroundTask
import re

import wtforms
from starlette.endpoints import HTTPEndpoint
from starlette.responses import JSONResponse, RedirectResponse, Response
from wtforms import validators as wtvalidators, widgets as wtwidgets
from wtforms.fields import html5 as h5fields

import otree
import otree.bots.browser
import otree.channels.utils as channel_utils
import otree.common
import otree.models
import otree.views.cbv
from otree import export
from otree import settings
from otree.common import (
    get_models_module,
    DebugTable,
    AUTH_COOKIE_NAME,
    AUTH_COOKIE_VALUE,
    participants_with_updated_presence_icons,
)
from otree.constants import ADVANCE_SLOWEST_BATCH_SIZE
from otree.currency import RealWorldCurrency
from otree.database import values_flat, save_sqlite_db, db
import otree.database
from otree.models import Session, Participant
from otree.session import SESSION_CONFIGS_DICT, SessionConfig
from otree.templating import get_template_name_if_exists
from otree.views.cbv import AdminSessionPage, AdminView
from . import cbv
from .cbv import enqueue_admin_message

validators_required = [wtvalidators.InputRequired()]


def mk_participant_code_urls(session: Session, request):

    p_codes = values_flat(session.pp_set.order_by('id_in_session'), Participant.code)
    participant_urls = []
    for code in p_codes:
        rel_url = otree.common.participant_start_url(code)
        url = request.base_url.replace(path=rel_url)
        participant_urls.append(url)
    return participant_urls


def pretty_name(name):
    """Converts 'first_name' to 'first name'"""
    if not name:
        return ''
    return name.replace('_', ' ')


class CreateSessionForm(wtforms.Form):
    session_configs = SESSION_CONFIGS_DICT.values()
    session_config_choices = [(s['name'], s['display_name']) for s in session_configs]

    session_config = wtforms.SelectField(
        choices=session_config_choices,
        validators=validators_required,
        render_kw=dict({'class': 'form-select'}),
    )

    num_participants = wtforms.IntegerField(
        validators=[wtvalidators.DataRequired(), wtvalidators.NumberRange(min=1)],
        render_kw={'autofocus': True, 'class': 'form-control w-auto'},
    )

    # too much weirdness with BooleanField and 'y'
    # so we render manually
    # it's a booleanfield so its default value will be 'y',
    # but it's a hidden widget that we are passing to the server
    # through .serializeArray, so we need to filter out
    is_mturk = wtforms.BooleanField()
    room_name = wtforms.StringField(widget=wtwidgets.HiddenInput())

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.is_mturk.object_data:
            label = "Number of MTurk workers (assignments)"
            description = (
                'Since workers can return an assignment or drop out, '
                'some "spare" participants will be created: '
                f'the oTree session will have {settings.MTURK_NUM_PARTICIPANTS_MULTIPLE} '
                'times more participant objects than the number you enter here.'
            )
        else:
            label = "Number of participants"
            description = ''

        self.num_participants.label = label
        self.num_participants.description = description

    def validate(self):
        if not super().validate():
            return False

        config = SESSION_CONFIGS_DICT[self.session_config.data]
        lcm = config.get_lcm()
        if self.num_participants.data % lcm:
            self.num_participants.errors.append(
                'Please enter a valid number of participants.'
            )
        return not bool(self.errors)


class CreateSession(cbv.AdminView):
    template_name = 'otree/CreateSession.html'
    url_pattern = '/create_session'

    def get_form(self):
        # need to pass is_mturk because it uses a different label.
        return CreateSessionForm(
            is_mturk=bool(self.request.query_params.get('is_mturk'))
        )

    def get_context_data(self, **kwargs):
        x = super().get_context_data(
            configs=SESSION_CONFIGS_DICT.values(),
            # splinter makes request.GET.get('mturk') == ['1\\']
            # no idea why
            # so just see if it's non-empty
            **kwargs,
        )
        return x


class SessionSplitScreen(AdminSessionPage):
    '''Launch the session in fullscreen mode
    only used in demo mode
    '''

    def vars_for_template(self):
        participant_urls = [
            self.request.base_url.replace(path=participant._start_url())
            for participant in self.session.get_participants()
        ]

        return dict(session=self.session, participant_urls=participant_urls)


class SessionDemoGridView(AdminSessionPage):
    '''Launch the session in fullscreen mode
    only used in demo mode
    '''

    def vars_for_template(self):
        participant_urls = [
            self.request.base_url.replace(path=participant._start_url())
            for participant in self.session.get_participants()
        ]

        items_per_row = math.ceil(len(participant_urls) / 2)

        return dict(
            session=self.session,
            participant_urls=participant_urls,
            items_per_row=items_per_row,
        )


class SessionStartLinks(AdminSessionPage):
    def inner_dispatch(self, request):
        self.session = db.get_or_404(Session, code=request.path_params['code'])
        room = self.session.get_room()
        if room:
            return self.redirect('RoomWithSession', room_name=room.name)
        return super().inner_dispatch(request)

    def vars_for_template(self):
        session = self.session

        participant_urls = mk_participant_code_urls(session, self.request)

        context = dict(
            use_browser_bots=session.use_browser_bots,
            participant_urls=participant_urls,
        )

        from otree.asgi import reverse

        anonymous_url = self.request.base_url.replace(
            path=reverse('SessionWideLink', anonymous_code=session._anonymous_code)
        )

        num_participants = len(participant_urls)

        context.update(
            anonymous_url=anonymous_url,
            num_participants=num_participants,
            splitscreen_mode_on=num_participants <= SPLITSCREEN_UPPER_LIMIT,
            grid_mode_on=GRID_LOWER_LIMIT <= num_participants <= GRID_UPPER_LIMIT,
            SPLITSCREEN_UPPER_LIMIT=SPLITSCREEN_UPPER_LIMIT,
            GRID_LOWER_LIMIT=GRID_LOWER_LIMIT,
            GRID_UPPER_LIMIT=GRID_UPPER_LIMIT,
        )

        return context


SPLITSCREEN_UPPER_LIMIT = 6
GRID_LOWER_LIMIT = 3
GRID_UPPER_LIMIT = 12


class SessionEditPropertiesForm(wtforms.Form):
    participation_fee = wtforms.DecimalField()
    real_world_currency_per_point = wtforms.DecimalField(places=6)
    label = wtforms.StringField()
    comment = wtforms.TextAreaField(render_kw=dict(rows='3', cols='40'))

    field_names = [
        'participation_fee',
        'real_world_currency_per_point',
        'label',
        'comment',
    ]


class SessionEditProperties(AdminSessionPage):

    form_class = SessionEditPropertiesForm

    def get_form(self):
        session = self.session
        config = session.config

        form = SessionEditPropertiesForm(
            data=dict(
                participation_fee=config['participation_fee'],
                real_world_currency_per_point=config['real_world_currency_per_point'],
                label=session.label,
                comment=session.comment,
            )
        )
        if session.mturk_HITId:
            form.participation_fee.widget = wtwidgets.HiddenInput()
        return form

    def form_valid(self, form):
        session = self.session
        session.label = form.label.data
        session.comment = form.comment.data

        participation_fee = form.participation_fee.data
        # convert to float because that's the usual type for rwc_per_point
        # it shouldn't behave differently after you edit session properties
        rwc_per_point = float(form.real_world_currency_per_point.data)

        config = session.config.copy()
        if participation_fee is not None:
            config['participation_fee'] = RealWorldCurrency(participation_fee)
        if rwc_per_point is not None:
            config['real_world_currency_per_point'] = rwc_per_point
        # need to do this to get SQLAlchemy to detect a change
        session.config = config
        enqueue_admin_message('success', 'Properties have been updated')
        return self.redirect('SessionEditProperties', code=session.code)


class SessionPayments(AdminSessionPage):
    def vars_for_template(self):
        session = self.session
        participants = session.get_participants()
        total_payments = 0.0
        mean_payment = 0.0
        if participants:
            total_payments = sum(
                pp.payoff_plus_participation_fee() for pp in participants
            )
            mean_payment = total_payments / len(participants)

        return dict(
            participants=participants,
            total_payments=total_payments,
            mean_payment=mean_payment,
            participation_fee=session.config['participation_fee'],
        )


class SessionDataAjax(AdminSessionPage):
    url_pattern = r"/session_data/{code}"

    async def get(self, request, code):
        import asyncio
        from otree.database import session_scope
        from otree.models import Session

        def get_data_in_thread():
            # Run blocking data export in background thread
            with session_scope():
                # Reload session in this thread's context
                session = Session.objects_get(code=code)
                return list(export.get_rows_for_data_tab(session))

        loop = asyncio.get_event_loop()
        rows = await loop.run_in_executor(None, get_data_in_thread)
        return JSONResponse(rows)


def break_on_underscores(text_list):
    # return text_list
    return [text.replace('_', ' ') for text in text_list]


class SessionData(AdminSessionPage):
    def vars_for_template(self):
        session = self.session

        tables = []
        app_names_by_subsession = []
        round_numbers_by_subsession = []
        table_index = 0
        for app_name in session.config['app_sequence']:
            models_module = get_models_module(app_name)
            num_rounds = models_module.Subsession.objects_filter(
                session=session
            ).count()
            pfields, gfields, sfields = export.get_fields_for_data_tab(app_name)

            for round_number in range(1, num_rounds + 1):
                # Build headers with indexes for sortability
                headers_with_indexes = []
                field_index = 0

                for field_name, prefix in [
                    *[(col['display'], col['prefix']) for col in export.DATA_TAB_FIXED_COLUMNS],
                    *[(f, '') for f in pfields],
                    *[(f, 'group.') for f in gfields],
                    *[(f, 'subsession.') for f in sfields],
                ]:
                    headers_with_indexes.append(
                        {
                            'name': field_name,
                            'display': break_on_underscores([field_name])[0],
                            'index': field_index,
                            'prefix': prefix,
                            'table_id': table_index,
                        }
                    )
                    field_index += 1

                tables.append(dict(headers=headers_with_indexes))
                app_names_by_subsession.append(app_name)
                round_numbers_by_subsession.append(round_number)
                table_index += 1
        return dict(
            tables=tables,
            app_names_by_subsession=app_names_by_subsession,
            round_numbers_by_subsession=round_numbers_by_subsession,
            GROUP_COL_INDEX=export.GROUP_COL_INDEX,
            DATA_EXPORT_HASH=otree.common.DATA_EXPORT_HASH,
        )


_FIELDS_AND_TITLES = dict(
    id_in_session='ID',
    code='Code',
    label='Label',
    _current_page_of_total='Progress',
    _current_app_name='App',
    _round_number='Round',
    _current_page_name='PageName',
    _presence='Waiting',
    _monitor_note_json='WaitsFor',
    _last_page_timestamp='Since',
)

_SORTABLE_FIELDS = [
    'id_in_session',
    'label',
    '_current_page_of_total',
    '_presence',
    '_monitor_note_json',
    '_last_page_timestamp',
]


class SessionMonitor(AdminSessionPage):
    def vars_for_template(self):
        field_names = export.get_fields_for_monitor()
        assert field_names == list(_FIELDS_AND_TITLES.keys())

        headers = []
        for field, title in _FIELDS_AND_TITLES.items():
            if field in _SORTABLE_FIELDS:
                title = (
                    f'<a href="#" data-sortable="true" data-field="{field}">{title}</a>'
                )
            headers.append(title)

        return dict(
            socket_url=channel_utils.session_monitor_path(self.session.code),
            ADVANCE_SLOWEST_BATCH_SIZE=ADVANCE_SLOWEST_BATCH_SIZE,
            headers=headers,
        )


class SessionMonitorAskForPresenceIcons(AdminSessionPage):
    # This uses AJAX with requests every N seconds.
    # The reason we use pull instead of push is that push
    # can result in a lot of extra traffic and flickering,
    # since when people navigate between pages their connection status
    # changes quickly. Better to batch it.
    # Also, we don't need to see presence icons update instantly.

    url_pattern = r"/presence_icons/{code}"

    def get(self, request, code):
        from otree import export

        session = self.session = db.get_or_404(Session, code=code)
        ids = participants_with_updated_presence_icons[session.id]
        if ids:
            participants = list(
                Participant.objects_filter(
                    Participant.id.in_(list(ids)),
                )
            )
            if participants:
                channel_utils.sync_group_send(
                    group=channel_utils.session_monitor_group_name(code),
                    # 2025-10-14: why do we get the full data when we are just supposed to get presence icons?
                    data=dict(rows=export.get_rows_for_monitor(participants)),
                )
            ids.clear()

        return JSONResponse({})


class SessionDescription(AdminSessionPage):
    def vars_for_template(self):
        return dict(config=SessionConfig(self.session.config))


class AdminReportForm(wtforms.Form):
    app_name = wtforms.SelectField(
        render_kw={'class': 'form-control'},
    )
    # use h5fields to get type='number' (but otree hides the spinner)
    round_number = h5fields.IntegerField(
        validators=[wtvalidators.Optional(), wtvalidators.NumberRange(min=1)],
        render_kw={'autofocus': True, 'class': 'form-control'},
    )

    def __init__(self, *args, session, **kwargs):
        '''we don't validate input it because we don't show the user
        an error. just coerce it to something right'''

        self.session = session
        admin_report_apps = self.session._admin_report_apps()
        num_rounds_list = self.session._admin_report_num_rounds_list()
        self.rounds_per_app = dict(zip(admin_report_apps, num_rounds_list))

        data = kwargs['data']
        # can't use setdefault because the key will always exist even if the
        # fields were empty.
        # str default value is '',
        # and int default value is None
        if not data.get('app_name'):
            data['app_name'] = admin_report_apps[0]
        rounds_in_this_app = self.rounds_per_app[data['app_name']]
        # use 0 so that we can bump it up in the next line
        round_number = int(data.get('round_number', 0))
        if not 1 <= round_number <= rounds_in_this_app:
            data['round_number'] = rounds_in_this_app

        super().__init__(*args, **kwargs)

        app_name_choices = []
        for app_name in admin_report_apps:
            label = f'{app_name} ({self.rounds_per_app[app_name]} rounds)'
            app_name_choices.append((app_name, label))
        self.app_name.choices = app_name_choices


class AdminReport(AdminSessionPage):
    def get_form(self):
        form = AdminReportForm(
            data=dict(self.request.query_params), session=self.session
        )
        form.validate()
        return form

    def get_context_data(self, **kwargs):
        form = kwargs['form']
        app_name = form.app_name.data
        models_module = get_models_module(app_name)
        subsession = models_module.Subsession.objects_get(
            session=self.session, round_number=form.round_number.data
        )

        target = subsession.get_user_defined_target()
        func = getattr(target, 'vars_for_admin_report', None)
        if func:
            vars_for_admin_report = func(subsession)
        else:
            vars_for_admin_report = {}

        self.debug_tables = [
            DebugTable(
                title='vars_for_admin_report', rows=vars_for_admin_report.items()
            )
        ]

        # todo: i think app_label and app_name are the same always
        app_label = subsession.get_folder_name()
        user_template = get_template_name_if_exists(
            [f'{app_label}/admin_report.html', f'{app_label}/AdminReport.html']
        )

        Constants = otree.common.get_constants(app_name)
        context = super().get_context_data(
            subsession=subsession,
            user_template=user_template,
            **kwargs,
        )
        context[Constants.__name__] = Constants
        # it's passed by parent class
        assert 'session' in context

        # this should take priority, in the event of a clash between
        # a user-defined var and a built-in one
        context.update(vars_for_admin_report)
        return context


def get_json_from_pypi() -> dict:
    # import only if we need it
    import urllib.request

    try:
        f = urllib.request.urlopen('https://pypi.python.org/pypi/otree/json', timeout=5)
        return json.loads(f.read().decode('utf-8'))
    except:
        return {'releases': []}


def get_installed_and_pypi_version() -> dict:
    '''return a dict because it needs to be json serialized for the AJAX
    response'''
    # need to import it so it can be patched outside

    semver_re = re.compile(r'^(\d+)\.(\d+)\.(\d+)$')

    installed_dotted = otree.__version__

    data = get_json_from_pypi()

    releases = data['releases']
    newest_tuple = [0, 0, 0]
    newest_dotted = ''
    for release in releases:
        release_match = semver_re.match(release)
        if release_match:
            release_tuple = [int(n) for n in release_match.groups()]
            if release_tuple > newest_tuple:
                newest_tuple = release_tuple
                newest_dotted = release
    return dict(newest=newest_dotted, installed=installed_dotted)


class ServerCheck(AdminView):
    url_pattern = '/server_check'

    def get_context_data(self, **kwargs):
        backend_name = otree.database.engine.url.get_backend_name()
        is_postgres = 'postgres' in backend_name.lower()
        return super().get_context_data(
            debug=settings.DEBUG,
            auth_level=settings.AUTH_LEVEL,
            auth_level_ok=settings.AUTH_LEVEL in {'DEMO', 'STUDY'},
            pypi_results=get_installed_and_pypi_version(),
            is_postgres=is_postgres,
            backend_name=backend_name,
            **kwargs,
        )


class AdvanceSession(AdminView):
    url_pattern = '/AdvanceSession/{code}'

    def post(self, request, code):
        session = db.get_or_404(Session, code=code)
        # background task because it makes http requests,
        # so it will need its own lock.
        post_data = self.get_post_data()
        pcode = post_data.get('selected_participant')

        if pcode:
            pp = db.get_or_404(Participant, code=pcode)
            session.advance_selected_participants([pp])
        else:
            mode = post_data['advancemode']
            session.advance_last_place_participants(visited_only=mode == 'visited_only')
        # task = BackgroundTask(session.advance_last_place_participants)
        return Response('ok')


class Sessions(AdminView):
    url_pattern = '/sessions'

    def vars_for_template(self):
        is_archive = bool(self.request.query_params.get('archived'))
        sessions = (
            Session.objects_filter(is_demo=False, archived=is_archive)
            .order_by(Session.id.desc())
            .all()
        )
        return dict(
            is_archive=is_archive,
            sessions=sessions,
            archived_sessions_exist=Session.objects_exists(archived=True),
        )


class ToggleArchivedSessions(AdminView):
    url_pattern = '/ToggleArchivedSessions'

    def post(self, request):
        post_data = self.get_post_data()
        code_list = post_data.getlist('session')
        for session in Session.objects_filter(Session.code.in_(code_list)):
            session.archived = not session.archived
        return self.redirect('Sessions')


class SaveDB(HTTPEndpoint):
    url_pattern = '/SaveDB'

    def post(self, request):
        import sys
        import os

        # prevent unauthorized requests
        if 'devserver_inner' in sys.argv:
            # very fast, ~0.05s
            save_sqlite_db()
        return Response(str(os.getpid()))


class LoginForm(wtforms.Form):
    username = wtforms.StringField()
    password = wtforms.StringField()

    def validate(self):
        if not super().validate():
            return False

        if (
            self.username.data == settings.ADMIN_USERNAME
            and self.password.data == settings.ADMIN_PASSWORD
        ):
            return True
        self.password.errors.append('Login failed')
        return False


class Login(AdminView):
    url_pattern = '/login'
    form_class = LoginForm

    def vars_for_template(self):
        warnings = []
        for setting in ['ADMIN_USERNAME', 'ADMIN_PASSWORD']:
            if not getattr(settings, setting, None):
                warnings.append(f'{setting} is undefined')
        return dict(warnings=warnings)

    def form_valid(self, form):
        self.request.session[AUTH_COOKIE_NAME] = AUTH_COOKIE_VALUE
        return self.redirect('DemoIndex')


class Logout(HTTPEndpoint):
    url_pattern = '/logout'

    def get(self, request):
        del request.session[AUTH_COOKIE_NAME]
        return RedirectResponse(request.url_for('Login'), status_code=302)


class RedirectToDemo(HTTPEndpoint):
    url_name = '/'

    def get(self, request):
        return RedirectResponse('/demo', status_code=302)



from otree.api import *


doc = """
Your app description
"""


class C(BaseConstants):
    NAME_IN_URL = '{{ app_name }}'
    PLAYERS_PER_GROUP = None
    NUM_ROUNDS = 1


class Subsession(BaseSubsession):
    pass


class Group(BaseGroup):
    pass


class Player(BasePlayer):
    num_errors = models.IntegerField(initial=0)


class Trial(BaseTrial):
    word = models.StringField()
    solution = models.StringField()
    response = models.StringField()


def creating_session(subsession: Subsession):
    # you can delete this read_csv if you prefer
    # to generate the trial data in your python code.
    rows = read_csv(__name__ + '/trials.csv', Trial)

    for p in subsession.get_players():
        for row in rows:
            Trial.create(player=p, word=row['word'], solution=row['solution'])


# PAGES
class Task(Page):
    trial_model = Trial
    trial_stimulus_fields = ['word']
    trial_response_fields = ['response']

    @staticmethod
    def before_next_page(player: Player, timeout_happened):
        # calculate the user's score
        for trial in Trial.filter(player=player):
            if trial.response != trial.solution:
                player.num_errors += 1


class Results(Page):
    @staticmethod
    def vars_for_template(player: Player):
        # get the variable 'trials' so it can be displayed in the HTML template
        trials = Trial.filter(player=player)
        return dict(trials=trials)


page_sequence = [Task, Results]



from otree.api import *


doc = """
Your app description
"""


class C(BaseConstants):
    NAME_IN_URL = '{{ app_name }}'
    PLAYERS_PER_GROUP = None
    NUM_ROUNDS = 1


class Subsession(BaseSubsession):
    pass


class Group(BaseGroup):
    pass


class Player(BasePlayer):
    pass


# PAGES
class MyPage(Page):
    pass


class ResultsWaitPage(WaitPage):
    pass


class Results(Page):
    pass


page_sequence = [MyPage, ResultsWaitPage, Results]



# Copyright 2022 Christopher Wickens. All rights reserved.

import logging

import sqlalchemy
import sqlalchemy.orm
import sqlalchemy.pool
from sqlalchemy import Column, ForeignKey
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import (
    relationship,
)
from sqlalchemy.orm.exc import NoResultFound  # noqa
from sqlalchemy.sql import sqltypes as st
from sqlalchemy.sql.functions import func

import otree.common
from otree.channels import utils as channel_utils
from otree.common2 import json_dumps
from otree.database import ExtraModel, dbq
from otree.lookup import get_page_lookup
from otree.models import Participant

logger = logging.getLogger(__name__)


class BaseTrial(ExtraModel):
    # We should have a class for this because:
    # - makes it easier to enforce correctness
    # - we may need to add stuff to it in the future.
    # - less code to loop and create trials in the regular case
    # - if someone forgets to assign queue_positions, then all queue positions will be None
    #   and it will skip right past the trials page, which will be confusing.
    # - if doing infinite iteration, the trials will not all be created at the same time.
    #   therefore, it's hard to know what the previous queue_position was.
    #   need to store it somewhere, which is a pain.

    __abstract__ = True

    # can't use autoincrement=True because that only has an effect
    # if the field is part of the primary key.
    queue_position = Column(st.Integer)
    page_name = Column(st.String)

    @declared_attr
    def player_id(cls):
        app_name = cls.get_folder_name()
        # needs to be nullable so re-grouping can happen
        return Column(st.Integer, ForeignKey(f'{app_name}_player.id'), nullable=True)

    @declared_attr
    def player(cls):
        # back_populates=f'{cls.__name__}_set'
        return relationship(
            f'{cls.__module__}.Player',
        )

    @classmethod
    def create(cls, **kwargs):
        if 'queue_position' not in kwargs:
            global _incrementing_queue_position
            if _incrementing_queue_position is None:
                _incrementing_queue_position = (
                    dbq(func.max(cls.queue_position)).scalar() or 1
                )
            _incrementing_queue_position += 1
            kwargs['queue_position'] = _incrementing_queue_position
        page_class = kwargs.pop('page', None)
        if isinstance(page_class, type):
            kwargs['page_name'] = page_class.__name__
        return super().create(**kwargs)


# queue position doesn't need to always contain the global max, but it must consistently increment
# for a given player, even between server restarts (because you might create trials on the fly).
# it's OK if 1 app passes queue_position explicitly, and another app auto-increments.
# because those queue positions will never get compared to each other.
_incrementing_queue_position = None


async def trial_payload_function(participant_code, page_name, msg):

    try:
        participant = Participant.objects_get(code=participant_code)
    except NoResultFound:
        logger.warning(f'Participant not found: {participant_code}')
        return

    def send_error():
        """need to put it in a function, otherwise we get a warning
        that the coroutine wasn't awaited."""
        return _send_back(
            participant_code,
            participant._index_in_pages,
            dict(type='error'),
        )

    lookup = get_page_lookup(participant._session_code, participant._index_in_pages)
    app_name = lookup.app_name
    models_module = otree.common.get_models_module(app_name)
    PageClass = lookup.page_class
    if page_name != PageClass.__name__:
        logger.warning(
            f'Ignoring message from {participant_code} because '
            f'they are on page {PageClass.__name__}, not {page_name}.'
        )
        await send_error()

    player = models_module.Player.objects_get(
        round_number=lookup.round_number, participant=participant
    )
    Trial = PageClass.trial_model
    page_name = PageClass.__name__
    trial = get_current_trial(Trial, player, page_name)
    msg_type = msg['type']
    is_page_load = msg_type == 'load'
    resp = dict(is_page_load=is_page_load, type=msg_type)
    if trial and msg_type == 'response':
        if trial.id != msg['trial_id']:
            await send_error()
            msg = (
                "Trials: server and client are out of sync. "
                "Check if there were any errors earlier."
            )
            raise Exception(msg)
        response: dict = msg['response']
        if hasattr(PageClass, 'evaluate_trial'):
            try:
                feedback = PageClass.evaluate_trial(trial, response)
            except Exception:
                await send_error()
                raise
        else:
            server_fields = set(PageClass.trial_response_fields)
            client_fields = response.keys()
            server_only = server_fields - client_fields
            if server_only:
                await send_error()
                msg = (
                    "The following fields are in trial_response_fields, "
                    f"but were not received from sendTrialResponse: {server_only}. "
                    "If this key was indeed sent, make sure its value is non-null."
                )
                raise Exception(msg)
            client_only = client_fields - server_fields
            if client_only:
                await send_error()
                msg = (
                    "The following fields were sent from the sendTrialResponse, "
                    f"but are not in trial_response_fields: {client_only}"
                )
                raise Exception(msg)
            try:
                # need to do it this way rather than having an overridable evaluate_trial,
                # because it's a static method, so has no access to trial_response_fields.
                # we don't enforce non-null, because some trial fields may be optional or N/A,
                # e.g. reaction time if the user didn't react.
                for attr in PageClass.trial_response_fields:
                    setattr(trial, attr, response[attr])
            except Exception:
                await send_error()
                raise
            trial.queue_position = None
            feedback = {}
        resp.update(feedback=feedback)
        trial = get_current_trial(Trial, player, page_name)
    if trial:
        resp.update(trial=encode_trial(trial, PageClass.trial_stimulus_fields))
    else:
        resp.update(trial=None, completed_all_trials=True)
    progress = get_progress(Trial, player, page_name)
    progress.update(PageClass.trial_page_vars(player))
    resp.update(progress=progress)
    await _send_back(
        participant.code,
        participant._index_in_pages,
        resp,
    )


def encode_trial(trial, fields):
    fields = fields.copy()
    if 'id' not in fields:
        fields.append('id')
    return {attr: getattr(trial, attr) for attr in fields}


def get_all_trials(Trial, player, page_name: str):
    return Trial.objects_filter(_filter_by_page(Trial, page_name), player=player)


def _filter_by_page(Trial, page_name):
    return sqlalchemy.or_(Trial.page_name == page_name, Trial.page_name == None)


def get_progress(Trial, player, page_name: str) -> dict:
    trials = get_all_trials(Trial, player, page_name)
    total = trials.count()
    remaining = trials.filter(Trial.queue_position != None).count()

    return dict(
        numTrials=total,
        numTrialsRemaining=remaining,
        numTrialsCompleted=total - remaining,
    )


def get_current_trial(Trial, player, page_name):
    return (
        get_all_trials(Trial, player, page_name)
        .filter(Trial.queue_position != None)
        .order_by('queue_position')
        .first()
    )


async def _send_back(pcode, page_index, resp):
    '''separate function for easier patching'''

    group_name = channel_utils.trial_group(pcode, page_index)
    await channel_utils.group_send(
        group=group_name,
        data=resp,
    )



import json
import datetime
import time

from sqlalchemy import Column as Column, ForeignKey
from sqlalchemy.orm import relationship
from sqlalchemy.sql import sqltypes as st
from starlette.exceptions import HTTPException
import otree.common
import otree.database
from otree.common import random_chars_8, ADMIN_SECRET_CODE
import otree.constants
from otree.database import MixinVars, CurrencyType, NoResultFound
from otree.lookup import url_i_should_be_on, get_page_lookup
import otree.channels.utils as channel_utils
from otree import settings


class Participant(MixinVars, otree.database.SSPPGModel):
    __tablename__ = 'otree_participant'

    session_id = Column(st.Integer, ForeignKey('otree_session.id', ondelete='CASCADE'))
    # getting integrityerror, so trying passive_deletes
    # https://stackoverflow.com/questions/5033547/sqlalchemy-cascade-delete
    session = relationship("Session", back_populates="pp_set")

    label = Column(st.String(100), nullable=True)

    id_in_session = Column(st.Integer, nullable=True)

    payoff = Column(CurrencyType, default=0)

    # status could be 'finished', 'dropped out', etc. Empty string means active/in progress
    status = Column(st.String(50), default='', nullable=False)

    # better to use a string so that it doesn't become unofficial API
    time_started_utc = Column(st.String(50), nullable=True)
    mturk_assignment_id = Column(st.String(50), nullable=True)
    mturk_worker_id = Column(st.String(50), nullable=True)

    _index_in_pages = Column(st.Integer, default=0, index=True)

    def _numeric_label(self):
        """the human-readable version."""
        return 'P{}'.format(self.id_in_session)

    _monitor_note_json = Column(st.String(300), default='[]')

    code = Column(
        st.String(16),
        default=random_chars_8,
        # set non-nullable, until we make our CharField non-nullable
        nullable=False,
        # unique implies DB index
        unique=True,
        index=True,
    )

    # useful when we don't want to load the whole session just to get the code
    _session_code = Column(st.String(16))

    visited = Column(
        st.Boolean,
        default=False,
        index=True,
    )

    # stores when the page was first visited
    _last_page_timestamp = Column(st.Integer, nullable=True)

    _last_request_timestamp = Column(st.Integer, nullable=True)

    is_on_wait_page = Column(st.Boolean, default=False)

    # these are both for the admin
    # In the changelist, simply call these "page" and "app"
    _current_page_name = Column(st.String(200), nullable=True)
    _current_app_name = Column(st.String(200), nullable=True)

    # doesn't actually hold any data. just used because we need
    # a column in the monitor table for the presence icon,
    # and the code currently expects an attribute
    _presence = ''  # Column(st.String(1), default=True)

    # only to be displayed in the admin participants changelist
    _round_number = Column(st.Integer, nullable=True)

    _current_form_page_url = Column(st.String(500))

    _max_page_index = Column(
        st.Integer,
    )

    _SETATTR_NO_FIELD_HINT = ' You can define it in the PARTICIPANT_FIELDS setting.'

    _is_bot = Column(st.Boolean, default=False)
    # can't start with an underscore because used in template
    # can't end with underscore because it's a django field (fields.E001)
    is_browser_bot = Column(st.Boolean, default=False)

    _timeout_expiration_time = otree.database.FloatField()
    _timeout_page_index = Column(
        st.Integer,
    )

    _waitpage_is_connected = Column(st.Boolean, default=False)
    _waitpage_tab_hidden = Column(st.Boolean, default=False)
    _gbat_page_index = Column(
        st.Integer,
    )
    _gbat_grouped = Column(
        st.Boolean,
    )

    def set_label(self, label):
        if not label:
            return
        if len(label) > 100:
            raise HTTPException(
                404, f'participant_label is too long or malformed: {label}'
            )
        self.label = label

    def _current_page_of_total(self):
        # don't put 'pages' because that causes wrapping which takes more space
        # since it's longer than the header
        return [self._index_in_pages, self._max_page_index]

    # because variables used in templates can't start with an underscore
    def current_page_(self):
        return f'{self._index_in_pages}/{self._max_page_index}'

    def get_players(self):
        """Used to calculate payoffs"""
        lst = []
        app_sequence = self.session.config['app_sequence']
        for app in app_sequence:
            models_module = otree.common.get_models_module(app)
            players = models_module.Player.objects_filter(participant=self).order_by(
                'round_number'
            )
            lst.extend(list(players))
        return lst

    def get_player(self, app_name, round_number):
        models_module = otree.common.get_models_module(app_name)
        PlayerClass = getattr(models_module, 'Player')
        try:
            return PlayerClass.objects_get(participant=self, round_number=round_number)
        except NoResultFound:
            msg = f"No player found for the given app_name={repr(app_name)} and round_number={repr(round_number)}"
            raise Exception(msg) from None

    def _url_i_should_be_on(self):
        if not self.visited:
            return self._start_url()
        if self._index_in_pages <= self._max_page_index:
            return url_i_should_be_on(
                self.code, self._session_code, self._index_in_pages
            )
        return '/OutOfRangeNotification/' + self.code

    def _start_url(self):
        return otree.common.participant_start_url(self.code)

    def payoff_in_real_world_currency(self):
        return self.payoff.to_real_world_currency(self.session)

    def payoff_plus_participation_fee(self):
        return self.session._get_payoff_plus_participation_fee(self.payoff)

    def _get_current_player(self):
        lookup = get_page_lookup(self._session_code, self._index_in_pages)
        models_module = otree.common.get_models_module(lookup.app_name)
        PlayerClass = getattr(models_module, 'Player')
        return PlayerClass.objects_get(
            participant=self, round_number=lookup.round_number
        )

    def initialize(self, participant_label):
        """in a separate function so that we can call it individually,
        e.g. from advance_last_place_participants"""
        pp = self
        if pp._index_in_pages == 0:
            pp._index_in_pages = 1
            pp.visited = True

            # participant.label might already have been set
            if not pp.label:
                pp.set_label(participant_label)

            # use UTC because daylight savings is not abandoned yet in EU
            # if anything, they will switch to permanent CEST (UTC+2)
            pp.time_started_utc = str(datetime.datetime.utcnow())
            pp._last_page_timestamp = int(time.time())

            from otree import common2

            row = common2.TimeSpentRow(
                session_code=pp._session_code,
                participant_id_in_session=pp.id_in_session,
                participant_code=pp.code,
                page_index=0,
                app_name='',
                page_name='InitializeParticipant',
                epoch_time_completed=int(time.time()),
                round_number=1,
                timeout_happened=0,
                is_wait_page=0,
            )
            common2.write_row_to_page_buffer(row)

    def _update_monitor_table(self):
        from otree import export

        channel_utils.sync_group_send(
            group=channel_utils.session_monitor_group_name(self._session_code),
            data=dict(rows=export.get_rows_for_monitor([self])),
        )

    def _get_page_instance(self):
        if self._index_in_pages > self._max_page_index:
            return
        page = get_page_lookup(
            self._session_code, self._index_in_pages
        ).page_class.instantiate_without_request()
        page.set_attributes(self)
        return page

    def _submit_current_page(self):
        from otree.api import Page

        page = self._get_page_instance()
        if isinstance(page, Page):
            from starlette.datastructures import FormData

            page._form_data = FormData(
                {
                    otree.constants.admin_secret_code: ADMIN_SECRET_CODE,
                    otree.constants.timeout_happened: '1',
                }
            )
            page.post()

    def _visit_current_page(self):
        """
        we need the redirect logic because if it's from ensure_pages_visited,
        then the previous page
        was a wait page, so that will not handle redirects.
        """
        # don't need to handle more redirects than that.
        for i in range(5):
            page = self._get_page_instance()
            if not page:
                return
            # note-to-self
            # it's possible that the slowest user is on a wait page,
            # especially if their browser is closed.
            # because they were waiting for another user who then
            # advanced past the wait page, but they were never
            # advanced themselves.
            resp = page.get()
            if not str(resp.status_code).startswith('3'):
                return



import enum
from decimal import Decimal
from typing import Dict

import wtforms
import wtforms_sqlalchemy.orm
from sqlalchemy.sql import sqltypes as st
from sqlalchemy.types import Boolean
from wtforms import validators
from wtforms_sqlalchemy.orm import converts

import otree.common
import otree.constants
import otree.models
from otree import settings
from otree.currency import Currency, to_dec
from otree.database import CurrencyType, DecimalFieldType
from . import fields
from . import widgets
from . import multistring
from ..i18n import core_gettext


def model_form(ModelClass, obj, only):
    field_args = {}

    for name in only:
        model_field = getattr(ModelClass, name)
        field_props = model_field.form_props

        # fa = field_args
        fa = {
            'validators': [],
            'render_kw': {},
            # to match Django behavior
            'id': f'id_{name}',
        }
        for prop_name in ['label', 'multiple']:
            if prop_name in field_props:
                fa[prop_name] = field_props[prop_name]

        target = obj.get_user_defined_target()

        func = getattr(target, f'{name}_choices', None)
        has_choices = False
        if func:
            choices = func(obj)
            has_choices = True
        elif 'choices' in field_props:
            choices = field_props['choices']
            has_choices = True

        if has_choices:
            fa['choices'] = expand_choice_tuples(choices)

        if not has_choices and fa.get('multiple'):
            raise ValueError("MultiStringField must have choices defined")

        if not has_choices and type(model_field.type) in [
            st.Integer,
            st.Float,
            CurrencyType,
            DecimalFieldType,
        ]:

            func = getattr(target, f'{name}_min', None)
            if func:
                min = func(obj)
            else:
                # 0 is the default, not None
                min = field_props.get('min', 0)

            func = getattr(target, f'{name}_max', None)
            if func:
                max = func(obj)
            else:
                max = field_props.get('max')

            if [min, max] != [None, None]:
                fa['validators'].append(validators.NumberRange(min=min, max=max))
            if min is not None:
                fa['render_kw'].update(min=to_dec(min))
            if max is not None:
                fa['render_kw'].update(max=to_dec(max))

        if not field_props.get('blank'):
            fa['validators'].append(validators.InputRequired())

        fa['description'] = field_props.get('help_text')

        widget = field_props.get('widget')
        if widget:
            # actually we should deprecate passing widget instances since they are probably mutable
            if isinstance(widget, type):
                # wtforms expects widget instances
                widget = widget()
            fa['widget'] = widget
        field_args[name] = fa

    return wtforms_sqlalchemy.orm.model_form(
        model=ModelClass,
        base_class=ModelForm,
        only=only,
        converter=ModelConverter(),
        field_args=field_args,
    )


def get_form(instance, field_names, view, formdata):
    instance._is_frozen = False

    FormClass = model_form(type(instance), obj=instance, only=field_names)
    form = FormClass(formdata=formdata, obj=instance, view=view)
    # because only= does not preserve order, so we need to store this
    # so that {{ formfields }} tag can access it.
    form.field_names = field_names
    instance._is_frozen = True
    return form


class ModelConverter(wtforms_sqlalchemy.orm.ModelConverterBase):
    def __init__(self, extra_converters=None, use_mro=True):
        super().__init__(extra_converters, use_mro=use_mro)

    @classmethod
    def _string_common(cls, column, field_args, **extra):
        if isinstance(column.type.length, int) and column.type.length:
            field_args["validators"].append(validators.Length(max=column.type.length))

    @converts("String")  # includes Unicode
    def conv_String(self, field_args, **extra):
        self._string_common(field_args=field_args, **extra)
        return get_choices_field(field_args, FormDataTypes.str) or fields.StringField(
            **field_args
        )

    @converts("Text")
    def conv_Text(self, field_args, **extra):
        self._string_common(field_args=field_args, **extra)
        return fields.TextAreaField(**field_args)

    @converts("Boolean")
    def conv_Boolean(self, field_args, **extra):
        field_args.setdefault('widget', widgets.RadioSelect())
        if isinstance(field_args['widget'], widgets.Checkbox):
            return fields.CheckboxField(**field_args)
        fld = get_choices_field(field_args, FormDataTypes.bool)
        return fld

    @converts("Integer")  # includes BigInteger and SmallInteger
    def handle_integer_types(self, model, column, field_args, **extra):
        unsigned = getattr(column.type, "unsigned", False)
        if unsigned:
            field_args["validators"].append(validators.NumberRange(min=0))
        return get_choices_field(field_args, FormDataTypes.int) or fields.IntegerField(
            **field_args
        )

    @converts("Numeric")  # includes DECIMAL, Float/FLOAT, REAL, and DOUBLE
    def handle_decimal_types(self, field_args, **extra):
        return get_choices_field(field_args, FormDataTypes.float) or fields.FloatField(
            **field_args,
        )

    @converts("CurrencyType")
    def handle_currency(self, field_args, **extra):
        return get_choices_field(
            field_args, FormDataTypes.Currency
        ) or fields.CurrencyField(**field_args)

    @converts("DecimalFieldType")
    def handle_decimal_field(self, column, field_args, **extra):
        from otree.decimal import DecimalUnit

        unit = column.type.unit

        if 'choices' in field_args:
            choices = field_args['choices']
            if isinstance(choices[0], (list, tuple)):
                field_args['choices'] = [(Decimal(v), label) for (v, label) in choices]
            else:
                field_args['choices'] = [(Decimal(v), v) for v in choices]

            field_args['coerce'] = lambda value: unit(value)

            if field_args.pop('multiple', None):
                return multistring.MultiStringFormField(**field_args)

            widget = field_args.pop('widget', None)
            if widget:
                widget = type(widget)
            return {
                widgets.RadioSelect: fields.RadioField,
                widgets.RadioSelectHorizontal: fields.RadioFieldHorizontal,
                widgets.TextInput: fields.StringField,
                None: fields.DropdownField,
            }[widget](**field_args)
        else:
            return fields.DecimalFormField(unit=unit, **field_args)


def bool_from_form_value(val):
    # when using test clients, we might get a non-string type
    # javascript uses 'false'
    if val in [None, False, '', '0', 'False', 'false']:
        return False
    return True


class FormDataTypes(enum.Enum):
    bool = 'bool'
    float = 'float'
    int = 'int'
    Currency = 'currency'
    str = 'str'


coerce_functions = {
    FormDataTypes.bool: bool_from_form_value,
    FormDataTypes.int: int,
    FormDataTypes.float: float,
    FormDataTypes.Currency: Currency,
    FormDataTypes.str: str,
}


def get_choices_field(fa, datatype: FormDataTypes):
    # fa means field_args
    if datatype == FormDataTypes.bool:

        fa.setdefault(
            'choices', [(True, core_gettext('Yes')), (False, core_gettext('No'))]
        )
    if 'choices' in fa:
        if datatype == FormDataTypes.Currency:
            before = fa['choices']
            if isinstance(before[0], (list, tuple)):
                after = [(to_dec(v), label) for (v, label) in before]
            else:
                after = [(to_dec(v), Currency(v)) for v in before]
            fa['choices'] = after
        fa['coerce'] = coerce_functions[datatype]
        # have to remove it because WTForms doesn't recognize it.
        if fa.pop('multiple', None):
            return multistring.MultiStringFormField(**fa)
        widget = fa.pop('widget', None)
        if widget:
            widget = type(widget)
        return {
            widgets.RadioSelect: fields.RadioField,
            widgets.RadioSelectHorizontal: fields.RadioFieldHorizontal,
            widgets.TextInput: fields.StringField,
            None: fields.DropdownField,
        }[widget](**fa)
    elif fa.get('widget') and isinstance(
        fa['widget'], (widgets.RadioSelect, widgets.RadioSelectHorizontal)
    ):
        msg = f'Field uses a radio/select widget but no choices are defined'
        raise Exception(msg)


class ModelForm(wtforms.Form):
    class Meta:
        # take first 2 chars because otherwise chinese has problems
        locales = [settings.LANGUAGE_CODE_ISO[:2]]

    _fields: Dict[str, wtforms.fields.Field]
    non_field_error = None
    field_names = []

    def __init__(
        self,
        view,
        formdata=None,
        obj=None,
        prefix='',
        data=None,
        meta=None,
        **kwargs,
    ):
        self.view = view
        self.instance = obj

        super().__init__(
            formdata=formdata, obj=obj, prefix=prefix, data=data, meta=meta, **kwargs
        )

    def _get_method_from_page_or_model(self, method_name):
        for obj in [self.view, self.instance]:
            if hasattr(obj, method_name):
                meth = getattr(obj, method_name)
                if callable(meth):
                    return meth

    def validate(self):
        super_validates = super().validate()
        # note-to-self
        # even if super doesn't validate,
        # we don't return False right away, because we want to validate all fields
        # individually at once.
        # this is useful e.g. for showing error messages in the UI, and SubmissionMustFail(error_fields=...)
        # i think i remember that calculating form.errors was slow, so avoid if possible.
        fields_with_errors = [] if super_validates else list(self.errors)
        ModelClass = type(self.instance)
        for name, field in self._fields.items():
            if name in fields_with_errors:
                continue

            column = getattr(ModelClass, name)
            if (
                column.type == Boolean
                or isinstance(column.type, Boolean)
                and field.data is None
                and not column.form_props.get('blank')
            ):
                msg = otree.constants.field_required_msg
                field.errors.append(msg)

            error_string = self.instance.call_user_defined(
                f'{name}_error_message', field.data, missing_ok=True
            )
            if error_string:
                field.errors.append(error_string)

        if not self.errors and hasattr(self.view, 'error_message'):
            error = self.view.call_user_defined('error_message', self.data)
            if error:
                if isinstance(error, dict):
                    self.non_field_error = error.pop('__all__', None)
                    for k, v in error.items():
                        getattr(self, k).errors.append(v)
                else:
                    self.non_field_error = error
        return not bool(self.errors or self.non_field_error)

    @property
    def errors(self):
        errors = super().errors
        if self.non_field_error:
            errors['__all__'] = self.non_field_error
        return errors

    def __iter__(self):
        return (self[k] for k in self.field_names)


def expand_choice_tuples(choices):
    '''
    Don't need it while generating the form,
    since wtforms also accepts flat lists.
    actually we need it for multi-select field.
    '''
    if not choices:
        return None
    if not isinstance(choices[0], (list, tuple)):
        choices = [(value, value) for value in choices]
    return choices



import asyncio
import os
import json

from starlette.concurrency import run_in_threadpool
from starlette.datastructures import FormData
from starlette.endpoints import HTTPEndpoint
from starlette.requests import Request
from starlette.responses import Response, RedirectResponse

from otree import settings
from otree.common import CSRF_TOKEN_NAME, AUTH_COOKIE_NAME, AUTH_COOKIE_VALUE, rng
from otree.database import db
from otree.models import Session
from otree.templating import render

ALWAYS_UNRESTRICTED = 'ALWAYS_UNRESTRICTED'
UNRESTRICTED_IN_DEMO_MODE = 'UNRESTRICTED_IN_DEMO_MODE'

_admin_message_queue = []


def enqueue_admin_message(bootstrap_alert_type, msg):
    _admin_message_queue.append(dict(alert_type=bootstrap_alert_type, msg=msg))


ADMIN_EMOJIS = {
    'SessionDescription': '📜',
    'SessionMonitor': '👓',
    'SessionData': '🔎',
    'SessionPayments': '💰',
    'SessionStartLinks': '🔗',
    'SessionSplitScreen': '◫',
    'SessionDemoGridView': '⊞',
    'SessionEditProperties': '✏️',
    'CreateDemoSession': '🔄',  # '➕',
    'RoomWithoutSession': '🚪',
    'RoomWithSession': '🚪',
}


class AdminView(HTTPEndpoint):
    csrf_exempt = False
    request: Request
    _requires_login = False
    form_class = None
    _form_data = None

    def _is_unauthorized(self):
        return self._requires_login and not self._is_logged_in()

    def _is_logged_in(self):
        return self.request.session.get(AUTH_COOKIE_NAME) == AUTH_COOKIE_VALUE

    async def dispatch(self) -> None:
        self.request = request = Request(self.scope, receive=self.receive)
        if request.method.lower() == 'post':
            # better done from async
            self._form_data = await self.request.form()
        response = self.inner_dispatch(request)
        # Handle both sync and async view methods
        if asyncio.iscoroutine(response):
            response = await response
        await response(self.scope, self.receive, self.send)

    def get_post_data(self) -> FormData:
        return self._form_data

    def inner_dispatch(self, request: Request):
        if self._is_unauthorized():
            return self.redirect('Login')
        resp = self.intercept_dispatch(**request.path_params)
        if resp:
            return resp
        if request.method.lower() == 'post':
            if not (self.csrf_exempt or os.getenv('OTREE_SKIP_CSRF')):
                post_data = self.get_post_data()
                form_token = post_data.get(CSRF_TOKEN_NAME)
                cookie = request.session.get(CSRF_TOKEN_NAME)
                if not form_token or not cookie or (form_token != cookie):
                    return Response(status_code=403, content='CSRF verification failed')

            return self.post(request, **request.path_params)
        return self.get(request, **request.path_params)

    def intercept_dispatch(self, **kwargs):
        pass

    def get(self, request, **kwargs):
        context = {}
        form = self.get_form()
        if form:
            context['form'] = form
        context = self.get_context_data(**context)
        return self.render_to_response(context)

    def get_form(self):
        if self.form_class:
            return self.form_class()

    def post(self, request, *args, **kwargs):
        form = self.form_class(formdata=self.get_post_data())
        if form.validate():
            return self.form_valid(form)
        return self.form_invalid(form)

    def form_valid(self, form):
        return self.redirect(self.get_success_url())

    def form_invalid(self, form):
        context = self.get_context_data(form=form)
        return self.render_to_response(context)

    def get_context_data(self, **kwargs):
        csrf_value = self.request.session.get(CSRF_TOKEN_NAME)
        if not csrf_value:
            csrf_value = str(rng.randint(10**9, 10**10))
            self.request.session[CSRF_TOKEN_NAME] = csrf_value
        token_input = (
            f'<input type="hidden" name="{CSRF_TOKEN_NAME}" value="{csrf_value}">'
        )
        kwargs[CSRF_TOKEN_NAME] = token_input
        # import functools
        current_page_name = self.__class__.__name__
        kwargs.update(
            view=self,
            admin_message_queue=_admin_message_queue,
            is_debug=settings.DEBUG,
            current_page_name=current_page_name,
            url_for=self.request.url_for,
            ADMIN_EMOJIS=ADMIN_EMOJIS,
            page_emoji=ADMIN_EMOJIS.get(current_page_name, ''),
            scrsplit='',
        )
        # vars_for_template has highest priority
        kwargs.update(self.vars_for_template())
        return kwargs

    def vars_for_template(self):
        '''
        simpler to use vars_for_template, but need to use get_context_data when:
        -   you need access to the context produced by the parent class,
            such as the form
        '''
        return {}

    def render_to_response(self, context):
        return render(self.get_template_name(), context)

    def get_success_url(self):
        return self.request.url

    def get_template_name(self):
        return 'otree/{}.html'.format(self.__class__.__name__)

    def redirect(self, url_name, **kwargs) -> RedirectResponse:
        return RedirectResponse(
            self.request.url_for(url_name, **kwargs), status_code=302
        )


class AdminSessionPage(AdminView):
    session: Session

    @classmethod
    def url_pattern(cls):
        return r"/%s/{code}" % cls.__name__

    def inner_dispatch(self, request):
        self.session = db.get_or_404(Session, code=request.path_params['code'])
        return super().inner_dispatch(request)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        scrsplit = self.request.query_params.get('scrsplit')
        scrsplit_param = 'scrsplit=true' if scrsplit else ''

        context.update(
            session=self.session,
            scrsplit=scrsplit,
            scrsplit_param=scrsplit_param,
            **kwargs,
        )

        return context


REST_KEY_NAME = 'OTREE_REST_KEY'
REST_KEY_HEADER = 'otree-rest-key'


def _HttpResponseForbidden(msg):
    return Response(msg, status_code=403)


def validate_rest_key(request):
    if settings.AUTH_LEVEL in ['DEMO', 'STUDY']:
        REST_KEY = os.getenv(REST_KEY_NAME)  # put it here for easy testing
        if not REST_KEY:
            return _HttpResponseForbidden(
                f'Env var {REST_KEY_NAME} must be defined to use REST API'
            )
        submitted_rest_key = request.headers.get(REST_KEY_HEADER)
        if not submitted_rest_key:
            return _HttpResponseForbidden(
                f'HTTP Request Header {REST_KEY_HEADER} is missing'
            )
        if REST_KEY != submitted_rest_key:
            return _HttpResponseForbidden(
                f'HTTP Request Header {REST_KEY_HEADER} is incorrect'
            )


class BaseRESTView(HTTPEndpoint):
    """
    async so that i can get request.json().
    inner_post should also call group_send, at there are complications with doing sync_group_send.

    """

    async def dispatch(self) -> None:
        request = self.request = Request(self.scope, receive=self.receive)
        # do the await out here in async function. need to compensate since we don't know
        # if there is a body (depending on the specific endpoint)
        body = await request.body()
        if body:
            payload = json.loads(body)
        else:
            payload = {}
        self._payload = payload
        response = self.inner_dispatch(request)
        await response(self.scope, self.receive, self.send)

    def inner_dispatch(self, request):
        error_response = validate_rest_key(request)
        if error_response:
            return error_response
        try:
            if request.method.lower() == 'post':
                return self.outer_post()
            return self.outer_get()
        except TypeError as exc:
            # assume it to be because of wrong kwargs
            return _HttpResponseForbidden(str(exc))

    def outer_get(self):
        # technically GET should be using the query params, not the JSON payload,
        # but it's just a pain to deal with the flimsy structure of query params
        # (untyped and awkward to pass lists as a multidict etc.)
        # wait until the API gets fleshed out more before I make the effort of switching
        # the only place we currently use json in GET is participant_labels, which is just an
        # optional param that filters down the data.
        return self.get(**self._payload)

    def outer_post(self):
        return self.post(**self._payload)

    # it's good to call these simply 'get' and 'post' because that makes the error message
    # simpler if they are called with wrong args (TypeError)
    def get(self, **kwargs):
        return Response('Method not allowed', status_code=405)

    def post(self, **kwargs):
        return Response('Method not allowed', status_code=405)
